#!/usr/bin/env python
# cardinal_pythonlib/spreadsheets.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Manipulate spreadsheets.**

Note:

- openpyxl is dreadfully slow. Its results are picklable, but not sensibly so
  (e.g. generating a >500Mb picklefile from a 12Mb spreadsheet.
- xlrd is much faster, but we can't pickle its results.

"""

import datetime
import decimal
from decimal import Decimal
import logging
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union
import unittest

from cardinal_pythonlib.datetimefunc import (
    coerce_to_pendulum,
    pendulum_to_datetime_stripping_tz,
)
from cardinal_pythonlib.progress import ActivityCounter
from cardinal_pythonlib.reprfunc import simple_repr
try:
    # noinspection PyPackageRequirements
    import xlrd
    # noinspection PyPackageRequirements
    from xlrd import Book
    # noinspection PyPackageRequirements
    from xlrd.sheet import Cell, Sheet
except ImportError:
    raise ImportError("You must install the 'xlrd' package.")

log = logging.getLogger(__name__)


# =============================================================================
# Consistency checks
# =============================================================================

def all_same(items: Iterable[Any]) -> bool:
    """
    Are all the items the same?

    https://stackoverflow.com/questions/3787908/python-determine-if-all-items-of-a-list-are-the-same-item

    ... though we will also allow "no items" to pass the test.
    """  # noqa
    return len(set(items)) <= 1


def values_by_attr(items: Sequence[Any], attr: str) -> List[Any]:
    """
    Returns the values of a given attribute for each of the ``items``.

    Args:
        items:
            Items to check
        attr:
            Name of attribute whose value should be taken across items.
    """
    return [getattr(item, attr) for item in items]


def attr_all_same(items: Sequence[Any], attr: str) -> bool:
    """
    Returns whether the value of an attribute is the same across a collection
    of items.

    Args:
        items:
            Items to check
        attr:
            Name of attribute whose value should be compared across items.
    """
    return all_same(values_by_attr(items, attr))


def check_attr_all_same(items: Sequence[Any],
                        attr: str,
                        id_attr: str = None,
                        fail_if_different: bool = True,
                        ignore_none: bool = False) -> None:
    """
    Checks if the value of an attribute is the same across a collection of
    items, and takes some action if not.

    Args:
        items:
            Items to check
        attr:
            Name of attribute whose value should be compared across items.
        id_attr:
            If the attributes are not all the same, use the value of this
            attribute from the first item to give some identifying context
            to the failure message.
        fail_if_different:
            If true, raises :exc:`ValueError` on failure; otherwise, prints a
            warning to the log.
        ignore_none:
            Ignore ``None`` values?
    """
    values = values_by_attr(items, attr)
    if ignore_none:
        values = [v for v in values if v is not None]
    if all_same(values):
        return
    # The rest of this function is about producing an error or a warning.
    first_item = items[0]
    if id_attr:
        identity = f"For {id_attr}={getattr(first_item, id_attr)!r}, "
    else:
        identity = ""
    msg = f"{identity}attribute {attr} is inconsistent: {values}"
    if fail_if_different:
        raise ValueError(msg)
    else:
        log.warning(msg)


def require_attr_all_same(items: Sequence[Any],
                          attr: str,
                          id_attr: str,
                          ignore_none: bool = False) -> None:
    """
    Raise if the ``attr`` attribute of each item in ``items`` is not the same.
    See :func:`check_attr_all_same`.
    """
    check_attr_all_same(items, attr, id_attr,
                        fail_if_different=True, ignore_none=ignore_none)


def prefer_attr_all_same(items: Sequence[Any],
                         attr: str,
                         id_attr: str,
                         ignore_none: bool = False) -> None:
    """
    Warn if the ``attr`` attribute of each item in ``items`` is not the same.
    See :func:`check_attr_all_same`.
    """
    check_attr_all_same(items, attr, id_attr,
                        fail_if_different=False, ignore_none=ignore_none)


# =============================================================================
# Spreadsheet operations: xlrd
# =============================================================================

def load_workbook(spreadsheet_filename: str) -> Book:
    """
    Load a workbook.
    """
    # Pickling creates massive files; skip it
    log.info(f"Loading: {spreadsheet_filename!r}...")
    book = xlrd.open_workbook(
        filename=spreadsheet_filename,
        on_demand=True  # may affect rather little, but we can try
    )
    log.info("... done")
    return book


def read_value_row(row: Sequence[Cell], colnum: int) -> Any:
    """
    Retrieves a value from a cell of a spreadsheet, given a row.

    AVOID: slower than index access (see :class:`SheetHolder`,
    :class:`RowHolder`).
    """
    return row[colnum].value


def read_int_row(row: Sequence[Cell], colnum: int) -> Optional[int]:
    """
    Reads an integer from a spreadsheet, given a row.

    AVOID: slower than index access (see :class:`SheetHolder`,
    :class:`RowHolder`).
    """
    v = read_value_row(row, colnum)
    if v is None:
        return None
    return int(v)


# =============================================================================
# Helper functions
# =============================================================================

def none_or_blank_string(x: Any) -> bool:
    """
    Is ``x`` either ``None`` or a string that is empty or contains nothing but
    whitespace?
    """
    if x is None:
        return True
    elif isinstance(x, str) and not x.strip():
        return True
    else:
        return False


def column_lettering(colnum: int) -> str:
    """
    Converts a zero-based column index into a spreadsheet-style column name
    (A[0] to Z[25], then AA[26] to AZ[51], etc). Basically, it's almost base
    26, but without a proper sense of zero (in that A is zero, but AA is 26).
    """
    assert colnum >= 0
    base = 26
    zero_char = ord("A")
    reversed_chars = ""
    while True:
        big, small = divmod(colnum, base)
        reversed_chars += chr(zero_char + small)
        if big == 0:
            break
        colnum = big - 1
    return reversed_chars[::-1]  # reverse again to get the final answer


def colnum_zb_from_alphacol(alphacol: str) -> int:
    """
    Reverses :func:`column_lettering`, generating a zero-based column index
    from an alphabetical name (A to Z, AA to AZ, etc.).
    """
    base = 26
    zero_char = ord("A")
    total = 0
    reversed_chars = alphacol[::-1]
    for pos, char in enumerate(reversed_chars):
        digit_value = ord(char) - zero_char  # e.g. 0 for A, 25 for Z
        assert 0 <= digit_value < base
        if pos > 0:
            digit_value += 1
        total += digit_value * pow(base, pos)
    return total


# =============================================================================
# SheetHolder
# =============================================================================

class SheetHolder(object):
    """
    Class to read from an Excel spreadsheet.
    """
    SHEET_NAME = ""  # may be overridden
    HEADER_ROW_ZERO_BASED = 0  # 0 is the first row (usually a header row)
    FIRST_DATA_ROW_ZERO_BASED = 1
    NULL_VALUES = [None, ""]

    BOOL_TRUE_VALUES_LOWERCASE = [1, "t", "true", "y", "yes"]
    BOOL_FALSE_VALUES_LOWERCASE = [0, "f", "false", "n", "no"]
    BOOL_UNKNOWN_VALUES_LOWERCASE = [None, "", "?", "not known", "unknown"]

    def __init__(self,
                 book: Book = None,
                 sheet_name: str = None,
                 sheet_index: int = None,
                 sheet: Sheet = None,
                 header_row_zero_based: int = None,
                 first_data_row_zero_based: int = None,
                 null_values: List[Any] = None,
                 bool_true_values_lowercase: List[Any] = None,
                 bool_false_values_lowercase: List[Any] = None,
                 bool_unknown_values_lowercase: List[Any] = None,
                 debug_max_rows_per_sheet: int = None) -> None:
        """
        There are two ways to specify the sheet:

        1.  Provide a workbook via ``book`` and...

            (a) a sheet number, or
            (b) a sheet name.

        2.  Provide a worksheet directly via ``sheet``.

        You can specify the following as ``_init__`` parameters or (via their
        capitalized versions) by subclassing:

        - sheet_name
        - header_row_zero_based
        - null_values
        - bool_true_values_lowercase
        - bool_false_values_lowercase
        - bool_unknown_values_lowercase

        Initialization parameters take priority over subclassed values.

        Args:
            book:
                Workbook, from which a worksheet should be selected.
            sheet_name:
                Name of a sheet to select from within ``book``.
            sheet_index:
                Index (zero-based) of a sheet to select from within ``book``.
            sheet:
                Worksheet, provided directly.
            header_row_zero_based:
                Row number (zero-based) of the header row.
            first_data_row_zero_based:
                Row number (zero-based) of the first row containing data.
            null_values:
                Values to treat as null (blank) values, converted to Python
                ``None``.
            bool_true_values_lowercase:
                Values to treat, by default, as ``True`` in Boolean columns.
            bool_false_values_lowercase:
                Values to treat, by default, as ``False`` in Boolean columns.
            bool_unknown_values_lowercase:
                Values to treat, by default, as missing/unknown in Boolean
                columns.
            debug_max_rows_per_sheet:
                Debugging option: the maximum number of data rows to
                process.
        """

        # Establish worksheet
        if book:
            assert sheet is None, (
                f"You specified 'book', so must not specify 'sheet', but "
                f"'sheet' was specified as: {sheet!r}"
            )
            if sheet_index is not None:
                self.sheet = book.sheet_by_index(sheet_index)
                self.sheet_description = (
                    f"book={book!r}, sheet_index={sheet_index!r}"
                )
            else:
                sheet_name = sheet_name or self.SHEET_NAME
                assert sheet_name, "Provide sheet_name or override SHEET_NAME"
                self.sheet = book.sheet_by_name(sheet_name)
                self.sheet_description = (
                    f"book={book!r}, sheet_name={sheet_name!r}"
                )
        else:
            assert sheet is not None, (
                "You didn't specify 'book', so must specify 'sheet', but it "
                "was None"
            )
            self.sheet = sheet
            self.sheet_description = f"sheet={sheet!r}"

        # Other parameters
        self.header_row_zero_based = (
            header_row_zero_based if header_row_zero_based is not None
            else self.HEADER_ROW_ZERO_BASED
        )
        self.first_data_row_zero_based = (
            first_data_row_zero_based if first_data_row_zero_based is not None
            else self.FIRST_DATA_ROW_ZERO_BASED
        )
        self.null_values = (
            null_values if null_values is not None
            else self.NULL_VALUES
        )
        self.bool_true_values_lowercase = (
            bool_true_values_lowercase
            if bool_true_values_lowercase is not None
            else self.BOOL_TRUE_VALUES_LOWERCASE
        )
        self.bool_false_values_lowercase = (
            bool_false_values_lowercase
            if bool_false_values_lowercase is not None
            else self.BOOL_FALSE_VALUES_LOWERCASE
        )
        self.bool_unknown_values_lowercase = (
            bool_unknown_values_lowercase
            if bool_unknown_values_lowercase is not None
            else self.BOOL_UNKNOWN_VALUES_LOWERCASE
        )
        self.debug_max_rows_per_sheet = debug_max_rows_per_sheet

        # Establish workbook
        self.book = self.sheet.book

        # Cached information
        self._headings = None  # type: Optional[List[str]]
        self._checked_headers = {}  # type: Dict[int, str]

    # -------------------------------------------------------------------------
    # Information
    # -------------------------------------------------------------------------

    def __str__(self) -> str:
        return (
            f"<Sheet {self.sheet_name!r}, specifed as "
            f"{self.sheet_description}>"
        )

    @property
    def n_rows(self) -> int:
        """
        Total number of rows.
        """
        return self.sheet.nrows

    @property
    def n_data_rows(self) -> int:
        """
        Total number of data rows (below any header row).
        """
        return self.n_rows - self.first_data_row_zero_based

    @property
    def sheet_name(self) -> str:
        """
        Name of the sheet within the workbook (file).
        """
        return self.sheet.name

    @property
    def headers(self) -> List[str]:
        """
        Returns all headings.
        """
        if self._headings is None:
            self._headings = [
                str(cell.value)
                for cell in self.sheet.row(self.header_row_zero_based)
            ]
        return self._headings

    @property
    def headings(self) -> List[str]:
        """
        Synonym for :data:`headers`.
        """
        return self.headers

    # -------------------------------------------------------------------------
    # Validation
    # -------------------------------------------------------------------------

    def _locinfo(self, row: int, col: int) -> str:
        """
        Location info for a cell, for errors.

        Args:
            row: zero-based row index
            col: zero-based column index
        """
        return (f" [sheet_name={self.sheet_name!r}, "
                f"row(1-based)={row+1}, "
                f"column(1-based)={col+1} ({column_lettering(col)})]")

    def ensure_header(self, col: int,
                      header: Union[str, Sequence[str]]) -> None:
        """
        Ensures that the header is correct for a specified column, or raise
        :exc:`ValueError`.

        You can specify a single correct heading or a sequence (e.g. list)
        of them.
        """
        if col in self._checked_headers:
            # Already checked.
            return
        headers = self.headers
        if col < 0 or col >= len(headers):
            max_col_idx = len(headers) - 1
            raise ValueError(
                f"Bad column index {col} ({column_lettering(col)}); "
                f"possible range is 0-{max_col_idx} (columns "
                f"{column_lettering(0)}-{column_lettering(max_col_idx)})"
            )
        v = headers[col]  # observed value
        self._checked_headers[col] = v  # cache for subsquent check
        if isinstance(header, str):  # single valid header
            if v == header:
                return  # good
        else:  # multiple values are OK
            if v in header:
                return  # good
        raise ValueError(
            f"Bad header: should be {header!r}, but was {v!r}" +
            self._locinfo(self.header_row_zero_based, col)
        )

    def ensure_heading(self, col: int,
                       heading: Union[str, Sequence[str]]) -> None:
        """
        Synonym for :meth:`ensure_header`.
        """
        return self.ensure_header(col, heading)

    # -------------------------------------------------------------------------
    # Reading
    # -------------------------------------------------------------------------

    def read_value(self, row: int, col: int,
                   check_header: Union[str, Sequence[str]] = None) -> Any:
        """
        Retrieves a value from a cell of a spreadsheet.

        Optionally, check that the heading for this column is correct (see
        :meth:`ensure_header`).
        """
        if check_header is not None:
            try:
                self.ensure_header(col, check_header)
            except ValueError as e:
                raise ValueError(
                    f"When reading row with zero-based index {row}: {e}")
        v = self.sheet.cell_value(row, col)
        if v in self.null_values:
            return None
        return v

    def read_datetime(self, row: int, col: int,
                      default: datetime.datetime = None,
                      check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.datetime]:
        """
        Reads a datetime from an Excel spreadsheet via xlrd.

        https://stackoverflow.com/questions/32430679/how-to-read-dates-using-xlrd
        """  # noqa
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        try:
            if isinstance(v, str):
                # Sometimes we get strings, like "1800-01-01". These are
                # outside the Excel range (1900 or 1904 onwards).
                p_datetime = coerce_to_pendulum(v)
                if p_datetime is None:
                    raise ValueError
                return pendulum_to_datetime_stripping_tz(p_datetime)
            else:
                # xlrd.xldate_as_tuple() converts an Excel number into a
                # tuple: (year, month, day, hour, minute, nearest_second)
                return datetime.datetime(
                    *xlrd.xldate_as_tuple(v, self.book.datemode)
                )
        except (TypeError, ValueError):
            raise ValueError(f"Bad date/time: {v!r}" + self._locinfo(row, col))

    def read_date(self, row: int, col: int,
                  default: datetime.date = None,
                  check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.date]:
        """
        Reads a date from an Excel spreadsheet

        https://stackoverflow.com/questions/32430679/how-to-read-dates-using-xlrd
        """
        dt = self.read_datetime(row, col, check_header=check_header)
        if dt:
            return dt.date()
        return default

    def read_int(
            self,
            row: int,
            col: int,
            default: int = None,
            check_header: Union[str, Sequence[str]] = None) -> Optional[int]:
        """
        Reads an integer from a spreadsheet.
        """
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        try:
            # - Now, the danger is that e.g. v == 7.6, which will cheerfully
            #   give int(v) == 7.
            # - [Note that int("7.6") will raise ValueError, whereas
            #   int(7.6) will succeed.]
            # - However, we want it to be an error to read a float/decimal as
            #   an integer accidentally, because otherwise we may lose data.
            # - Finally, sometimes str(v) is e.g. "92.0" when v is a perfectly
            #   valid integer, so we can't use 'if "." in str(v)' as a test.
            # - But comparing int(v) to float(v) works.
            iv = int(v)  # may raise
            if iv != float(v):  # this picks up non-integer values
                raise ValueError
            return iv
        except (TypeError, ValueError):
            raise ValueError(f"Bad int: {v!r}" + self._locinfo(row, col))

    def read_float(self, row: int, col: int,
                   default: float = None,
                   check_header: Union[str, Sequence[str]] = None) \
            -> Optional[float]:
        """
        Reads a float from the spreadsheet.
        """
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        try:
            return float(v)
        except (TypeError, ValueError):
            raise ValueError(f"Bad float: {v!r}" + self._locinfo(row, col))

    def read_decimal(
            self,
            row: int,
            col: int,
            default: Decimal = None,
            check_header: Union[str, Sequence[str]] = None,
            dp: int = None,
            rounding: str = decimal.ROUND_HALF_UP) -> Optional[Decimal]:
        """
        Reads a Decimal from the spreadsheet.

        If ``dp`` is not ``None``, force the result to a specified number of
        decimal places, using the specified rounding method.
        """
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        try:
            x = Decimal(str(v))
            # ... better than Decimal(v), which converts e.g. 7.4 to
            # Decimal('7.4000000000000003552713678800500929355621337890625')
        except (TypeError, decimal.InvalidOperation):
            raise ValueError(f"Bad decimal: {v!r}" + self._locinfo(row, col))
        if dp is not None:
            nplaces = Decimal(10) ** (-dp)
            x = x.quantize(exp=nplaces, rounding=rounding)
        return x

    def read_str(
            self,
            row: int,
            col: int,
            default: str = None,
            check_header: Union[str, Sequence[str]] = None) -> Optional[str]:
        """
        Reads a string from a spreadsheet.
        """
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        return str(v).strip()

    def read_str_int(self, row: int, col: int,
                     default: str = None,
                     check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        """
        Reads a string version of an integer. (This prevents e.g. "2" being
        read as a floating-point value of "2.0" then converted to a string.)
        """
        v_int = self.read_int(row, col, check_header=check_header)
        if v_int is None:
            return default
        return str(v_int)

    def read_str_nonfloat(self, row: int, col: int,
                          default: str = None,
                          check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        """
        Reads something that may be a string or numeric, but if it's numeric,
        it's integer (not float). (This prevents e.g. "2" being read as a
        floating-point value of "2.0" then converted to a string.)
        """
        v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(v):
            return default
        # See read_int() for more on the logic used here.
        try:
            fv = float(v)  # may raise
        except (TypeError, ValueError):
            # Not numeric
            return v  # return the string version
        # If we get here, v is numeric (e.g. "7" or "6.5" or 3.5).
        try:
            iv = int(v)
        except (TypeError, ValueError):
            # Numeric, but not integer
            raise ValueError
        if iv != fv:  # this picks up non-integer values
            raise ValueError
        return str(iv)  # string version of int

    def read_bool(self,
                  row: int,
                  col: int,
                  default: bool = None,
                  true_values_lowercase: List[Any] = None,
                  false_values_lowercase: List[Any] = None,
                  unknown_values_lowercase: List[Any] = None,
                  check_header: Union[str, Sequence[str]] = None) \
            -> Optional[bool]:
        """
        Reads a boolean value.
        """
        if true_values_lowercase is None:
            true_values_lowercase = self.bool_true_values_lowercase
        if false_values_lowercase is None:
            false_values_lowercase = self.bool_false_values_lowercase
        if unknown_values_lowercase is None:
            unknown_values_lowercase = self.bool_unknown_values_lowercase
        raw_v = self.read_value(row, col, check_header=check_header)
        if none_or_blank_string(raw_v):
            v = None
        else:
            try:
                v = int(raw_v)
            except (TypeError, ValueError):
                v = str(raw_v).lower()
        if v in true_values_lowercase:
            return True
        elif v in false_values_lowercase:
            return False
        elif v in unknown_values_lowercase:
            return default
        else:
            raise ValueError(f"Bad bool: {raw_v!r}" + self._locinfo(row, col))

    def read_none(self, row: int, col: int,
                  check_header: Union[str, Sequence[str]] = None) -> None:
        """
        Reads a value, and checks that it is a none/null value
        """
        v = self.read_value(row, col, check_header=check_header)
        if v is not None:
            raise ValueError(
                f"Value should be none/null but is {v!r}" +
                self._locinfo(row, col)
            )
        return None

    # -------------------------------------------------------------------------
    # Row generators
    # -------------------------------------------------------------------------

    def _setup_for_gen(self, with_counter: bool = True) \
            -> Tuple[int, Optional["ActivityCounter"]]:
        n_rows = self.sheet.nrows
        if with_counter:
            counter = ActivityCounter(
                activity="Reading row", n_total=self.n_data_rows)
        else:
            counter = None
        if self.debug_max_rows_per_sheet is not None:
            log.debug(
                f"Debug option: limiting to {self.debug_max_rows_per_sheet} "
                f"rows from spreadsheet {self}")
            end = min(n_rows, self.debug_max_rows_per_sheet + 1)
        else:
            end = n_rows
        return end, counter

    def gen_row_numbers_excluding_header_row(
            self, with_counter: bool = True) -> Iterable[int]:
        """
        Generates row numbers.

        xlrd uses 0-based numbering, so row 1 is the first beyond a header row.
        """
        end, counter = self._setup_for_gen(with_counter)
        for rownum in range(self.first_data_row_zero_based, end):
            if counter is not None:
                counter.tick()
            yield rownum

    def gen_rows_excluding_header_row(
            self, with_counter: bool = True) -> Iterable[Sequence[Cell]]:
        """
        Generates rows. AVOID; index-based access is faster.

        xlrd uses 0-based numbering, so row 1 is the first beyond a header row.
        """
        end, counter = self._setup_for_gen(with_counter)
        for index in range(self.first_data_row_zero_based, end):
            if counter is not None:
                counter.tick()
            yield self.sheet.row(index)


# =============================================================================
# RowHolder
# =============================================================================

class RowHolder(object):
    """
    Class to read from a single row of a spreadsheet.

    The intended use is to create something like a dataclass, but one that
    knows its spreadsheet structure. Like this:

    .. code-block:: python

        from cardinal_pythonlib.spreadsheets import RowHolder, SheetHolder

        class ReferralSheetHolder(SheetHolder):
            SHEET_NAME = "Patient Referrals 2018-19"

            def gen_referral_rows(self) -> Iterable["ReferralRow"]:
                for rownum in self.gen_row_numbers_excluding_header_row():
                    yield ReferralRow(self, rownum)

        class ReferralRow(RowHolder):
            def __init__(self, sheetholder: SheetHolder, row: int) -> None:
                super().__init__(sheetholder, row)

                self.inc_next_col()  # column 0: query period; ignore
                self.patient_id = self.str_int_pp()
                self.referral_id_within_patient = self.int_pp()
                self.age_at_referral_int = self.int_pp()
                self.ethnicity = self.str_pp()
                self.gender = self.str_pp(check_header="Gender")

        def import_referrals(book: Book) -> None:
            sheet = ReferralSheetHolder(book)
            for referral in sheet.gen_referral_rows():
                pass  # do something useful here

    """

    def __init__(self, sheetholder: SheetHolder, row: int) -> None:
        self.sheetholder = sheetholder
        self.row = row  # zero-based index of our row
        self._next_col = 0  # zero-based column index of the next column to read  # noqa

    # -------------------------------------------------------------------------
    # Information
    # -------------------------------------------------------------------------

    @property
    def sheet_name(self) -> str:
        return self.sheetholder.sheet_name

    def _get_relevant_attrs(self) -> List[str]:
        """
        Attributes added by the user, and row number.
        :return:
        """
        avoid = [
            "sheetholder",
            "row",
            "_next_col",
        ]
        user_attrs = [k for k in self.__dict__.keys() if k not in avoid]
        attrs = ["sheet_name", "row"] + user_attrs
        return attrs

    def __str__(self) -> str:
        return simple_repr(self, self._get_relevant_attrs())

    def __repr__(self) -> str:
        return simple_repr(self, ["sheetholder", "row"])

    @property
    def row_zero_based(self) -> int:
        """
        Zero-based row number.
        """
        return self.row

    @property
    def row_one_based(self) -> int:
        """
        One-based row number.
        """
        return self.row + 1

    # -------------------------------------------------------------------------
    # Checks
    # -------------------------------------------------------------------------
    # More commonly, this is done via the "read...()" functions.

    def ensure_header(self, col: int,
                      header: Union[str, Sequence[str]]) -> None:
        """
        Ensures the column has an appropriate heading value, or raises
        :exc:`ValueError`.
        """
        self.sheetholder.ensure_header(col, header)

    def ensure_heading(self, col: int,
                       header: Union[str, Sequence[str]]) -> None:
        """
        Synonym for :meth:`ensure_header`.
        """
        self.ensure_header(col, header)

    # -------------------------------------------------------------------------
    # Read operations, given a column number
    # -------------------------------------------------------------------------
    # Compare equivalents in SheetHolder.

    def read_value(self, col: int,
                   check_header: Union[str, Sequence[str]] = None) -> Any:
        return self.sheetholder.read_value(
            self.row, col, check_header=check_header)

    def read_datetime(self, col: int,
                      default: Any = None,
                      check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.date]:
        return self.sheetholder.read_datetime(
            self.row, col, default, check_header=check_header)

    def read_date(self, col: int,
                  default: Any = None,
                  check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.date]:
        return self.sheetholder.read_date(
            self.row, col, default, check_header=check_header)

    def read_int(self, col: int,
                 default: int = None,
                 check_header: Union[str, Sequence[str]] = None) \
            -> Optional[int]:
        return self.sheetholder.read_int(
            self.row, col, default, check_header=check_header)

    def read_float(self, col: int,
                   default: float = None,
                   check_header: Union[str, Sequence[str]] = None) \
            -> Optional[float]:
        return self.sheetholder.read_float(
            self.row, col, default, check_header=check_header)

    def read_decimal(
            self,
            col: int,
            default: Decimal = None,
            check_header: Union[str, Sequence[str]] = None,
            dp: int = None,
            rounding: str = decimal.ROUND_HALF_UP) -> Optional[Decimal]:
        return self.sheetholder.read_decimal(
            self.row,
            col,
            default,
            check_header=check_header,
            dp=dp,
            rounding=rounding
        )

    def read_str(
            self,
            col: int,
            default: str = None,
            check_header: Union[str, Sequence[str]] = None) -> Optional[str]:
        return self.sheetholder.read_str(
            self.row, col, default, check_header=check_header)

    def read_str_int(self, col: int,
                     default: str = None,
                     check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        return self.sheetholder.read_str_int(
            self.row, col, default, check_header=check_header)

    def read_str_nonfloat(self, col: int,
                          default: str = None,
                          check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        return self.sheetholder.read_str_nonfloat(
            self.row, col, default, check_header=check_header)

    def read_bool(self,
                  col: int,
                  default: bool = None,
                  true_values_lowercase: List[Any] = None,
                  false_values_lowercase: List[Any] = None,
                  unknown_values_lowercase: List[Any] = None,
                  check_header: Union[str, Sequence[str]] = None) \
            -> Optional[bool]:
        return self.sheetholder.read_bool(
            row=self.row,
            col=col,
            default=default,
            true_values_lowercase=true_values_lowercase,
            false_values_lowercase=false_values_lowercase,
            unknown_values_lowercase=unknown_values_lowercase,
            check_header=check_header
        )

    def read_none(self,
                  col: int,
                  check_header: Union[str, Sequence[str]] = None) -> None:
        return self.sheetholder.read_none(
            row=self.row,
            col=col,
            check_header=check_header
        )

    # -------------------------------------------------------------------------
    # Check the next column
    # -------------------------------------------------------------------------

    def ensure_next_col_header(self,
                               header: Union[str, Sequence[str]]) -> None:
        """
        Ensures the next column has an appropriate heading value, or raises
        :exc:`ValueError`.
        """
        self.ensure_header(self._next_col, header)

    def ensure_next_col_heading(self,
                                header: Union[str, Sequence[str]]) -> None:
        """
        Synonym for :meth:`ensure_next_col_header`.
        """
        self.ensure_next_col_header(header)

    # -------------------------------------------------------------------------
    # Read operations, incrementing the next column number automatically.
    # -------------------------------------------------------------------------
    # "pp" for "++" post-increment, like C.

    @property
    def next_col(self) -> int:
        """
        Returns the column number (0-based) that will be used by the next
        automatic read operation.
        """
        return self._next_col

    def set_next_col(self, col: int) -> None:
        """
        Resets the next column to be read automatically.
        """
        self._next_col = col

    def inc_next_col(self) -> None:
        """
        Increments the next column to be read.
        """
        self._next_col += 1

    def value_pp(self, check_header: Union[str, Sequence[str]] = None) -> Any:
        """
        Reads a value, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_value(self._next_col, check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def datetime_pp(self,
                    default: datetime.datetime = None,
                    check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.datetime]:
        """
        Reads a datetime, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_datetime(
                self._next_col, default, check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def date_pp(self,
                default: datetime.date = None,
                check_header: Union[str, Sequence[str]] = None) \
            -> Optional[datetime.date]:
        """
        Reads a date, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_date(self._next_col, default,
                               check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def int_pp(self,
               default: int = None,
               check_header: Union[str, Sequence[str]] = None) -> Optional[int]:
        """
        Reads an int, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_int(self._next_col, default,
                              check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def float_pp(self,
                 default: float = None,
                 check_header: Union[str, Sequence[str]] = None) \
            -> Optional[float]:
        """
        Reads a float, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_float(self._next_col, default,
                                check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def decimal_pp(self,
                   default: float = None,
                   check_header: Union[str, Sequence[str]] = None,
                   dp: int = None,
                   rounding: str = decimal.ROUND_HALF_UP) -> Optional[Decimal]:
        """
        Reads a Decimal, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_decimal(
                self._next_col,
                default=default,
                check_header=check_header,
                dp=dp,
                rounding=rounding
            )
        finally:
            self.inc_next_col()
        return v

    def str_pp(self,
               default: str = None,
               check_header: Union[str, Sequence[str]] = None) -> Optional[str]:
        """
        Reads a string, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_str(self._next_col, default,
                              check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def str_int_pp(self,
                   default: str = None,
                   check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        """
        Reads an integer as a string, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_str_int(
                self._next_col, default, check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def str_nonfloat_pp(self,
                        default: str = None,
                        check_header: Union[str, Sequence[str]] = None) \
            -> Optional[str]:
        """
        Reads something that may be a string or numeric, but if it's numeric,
        it's integer (not float). Then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_str_nonfloat(
                self._next_col, default, check_header=check_header)
        finally:
            self.inc_next_col()
        return v

    def bool_pp(
            self,
            default: bool = None,
            true_values_lowercase: List[Any] = None,
            false_values_lowercase: List[Any] = None,
            unknown_values_lowercase: List[Any] = None,
            check_header: Union[str, Sequence[str]] = None) -> Optional[bool]:
        """
        Reads a boolean value, then increments the "current" column.
        Optionally, checks that the header for this column is as expected.
        """
        try:
            v = self.read_bool(
                col=self._next_col,
                default=default,
                true_values_lowercase=true_values_lowercase,
                false_values_lowercase=false_values_lowercase,
                unknown_values_lowercase=unknown_values_lowercase,
                check_header=check_header
            )
        finally:
            self.inc_next_col()
        return v

    def none_pp(self, check_header: Union[str, Sequence[str]] = None) -> None:
        """
        Reads a null value, and ensures that it is null; then increments the
        "current" column. Optionally, checks that the header for this column is
        as expected.
        """
        try:
            self.read_none(col=self._next_col, check_header=check_header)
        finally:
            self.inc_next_col()
        return None


# =============================================================================
# Self-testing
# =============================================================================

class TestRoundingAndReversal(unittest.TestCase):

    def test_column_lettering(self) -> None:
        assert column_lettering(0) == "A"
        assert column_lettering(25) == "Z"
        assert column_lettering(26) == "AA"
        assert column_lettering(51) == "AZ"
        assert column_lettering(52) == "BA"
        for col_zb in range(200):
            alphacol = column_lettering(col_zb)
            assert colnum_zb_from_alphacol(alphacol) == col_zb


# =============================================================================
# Command-line entry point
# =============================================================================

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# cardinal_pythonlib/bulk_email/main.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

Command-line entry point for the simple bulk e-mail tool.

"""

# =============================================================================
# Imports
# =============================================================================

import argparse
from copy import deepcopy
from datetime import datetime
from email.utils import format_datetime as format_email_datetime
import logging
import os
import sys
from time import sleep
from typing import List, Optional, Set

# noinspection PyProtectedMember
from sqlalchemy.engine import create_engine, Engine
from sqlalchemy.orm.session import Session

from cardinal_pythonlib.bulk_email.constants import (
    DB_URL_ENVVAR,
    DEFAULT_TIME_BETWEEN_EMAILS_S,
    PASSWORD_OBSCURING_STRING,
)
from cardinal_pythonlib.bulk_email.models import (
    Base,
    Config,
    Content,
    Job,
    Recipient,
    SendAttempt,
)
from cardinal_pythonlib.email.sendmail import (
    CONTENT_TYPE_HTML,
    CONTENT_TYPE_TEXT,
    is_email_valid,
    send_email,
    STANDARD_SMTP_PORT,
    STANDARD_TLS_PORT,
    UTF8,
)
from cardinal_pythonlib.file_io import gen_lines_without_comments
from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger
from cardinal_pythonlib.sqlalchemy.session import get_safe_url_from_engine
from cardinal_pythonlib.sysops import EXIT_FAILURE, EXIT_SUCCESS

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

class Command:
    """
    Commands to this tool.
    """
    ADD = "add"
    CLEAR = "clear"
    CONFIGURE = "configure"
    CREATE_DB = "create_database_destroying_existing_data"
    INFO = "info"
    WORK = "work"


# =============================================================================
# Create database
# =============================================================================

def create_db(engine: Engine) -> None:
    """
    Creates the database.
    """
    log.info("Dropping existing tables...")
    Base.metadata.drop_all(engine)

    log.info("Creating tables...")
    Base.metadata.create_all(engine)

    log.info("Database created.")


# =============================================================================
# Configure server settings
# =============================================================================

def configure(session: Session,
              host: str,
              port: int,
              use_tls: bool,
              username: str,
              password: str,
              time_between_emails: float) -> None:
    """
    Updates the configuration from command-line parameters.
    """
    cfg = Config.get_or_create_config(session)

    cfg.host = host
    cfg.port = port
    cfg.use_tls = use_tls
    cfg.username = username
    cfg.password = password
    cfg.time_between_emails = time_between_emails

    session.commit()
    log.info("Configuration updated.")


# =============================================================================
# Add a job
# =============================================================================

def add_job(session: Session,
            recipients_filename: str,
            from_addr: str,  # avoid clash with "from" keyword
            reply_to_addr: str,
            subject: str,
            content_html_filename: Optional[str],
            content_text_filename: Optional[str],
            charset: str,
            timestamp_at_creation: bool) -> None:
    """
    Adds an e-mail sending job -- contents and recipients.
    """
    # -------------------------------------------------------------------------
    # Recipients
    # -------------------------------------------------------------------------

    log.info(f"Reading recipients from: {recipients_filename}")
    recipient_addresses = set()  # type: Set[str]
    recipients_failed = False
    for recipient_addr in gen_lines_without_comments(recipients_filename):
        if not is_email_valid(recipient_addr):
            log.error(f"Bad recipient e-mail: {recipient_addr}")
            recipients_failed = True
        else:
            recipient_addresses.add(recipient_addr)
    if recipients_failed:
        raise ValueError("Errors in recipient addresses")

    recipients = []  # type: List[Recipient]
    for recipient_addr in sorted(recipient_addresses):
        recipient = Recipient.fetch_or_add(recipient_addr, session)
        session.add(recipient)
        recipients.append(recipient)
    log.info(f"Read {len(recipients)} recipients.")

    # -------------------------------------------------------------------------
    # Contents
    # -------------------------------------------------------------------------

    log.info("Creating template...")
    if timestamp_at_creation:
        email_datetime = datetime.now()
    else:
        email_datetime = None  # timestamp at sending

    assert bool(content_text_filename) != bool(content_html_filename)
    if content_text_filename:
        filename = content_text_filename
        content_type = CONTENT_TYPE_TEXT
    else:
        filename = content_html_filename
        content_type = CONTENT_TYPE_HTML
    with open(filename) as f:
        body = f.read()

    content = Content(
        from_addr=from_addr,
        reply_to_addr=reply_to_addr,
        email_datetime=email_datetime,
        subject=subject,
        body=body,
        content_type=content_type,
        charset=charset
    )
    session.add(content)
    log.info("... created.")

    # -------------------------------------------------------------------------
    # Jobs
    # -------------------------------------------------------------------------

    log.info("Adding jobs...")
    for recipient in recipients:
        job = Job(
            recipient=recipient,
            content=content
        )
        session.add(job)
    log.info("... done.")


# =============================================================================
# Send pending emails
# =============================================================================

def work(session: Session,
         stop_on_error: bool) -> None:
    """
    Processes outstanding jobs until there are no more.
    """
    log.info("Processing pending jobs...")
    config = Config.get_current_config(session)
    sleep_time_s = config.time_between_emails
    for i, job in enumerate(Job.gen_jobs_to_be_done(session)):
        if i != 0:
            log.info(f"Sleeping for {sleep_time_s} s...")
            sleep(sleep_time_s)
        content = job.content
        recipient = job.recipient
        log.info(f"Sending to: {recipient.email}")
        now = datetime.now()
        if content.email_datetime:
            email_datetime = format_email_datetime(content.email_datetime)
        else:
            email_datetime = None  # "now"
        success, details = send_email(
            from_addr=content.from_addr,
            host=config.host,
            user=config.username,
            password=config.password,
            port=config.port,
            use_tls=config.use_tls,
            date=email_datetime,  # may be None for "now"
            sender="",
            reply_to=content.reply_to_addr,
            to=recipient.email,
            cc="",
            bcc="",
            subject=content.subject,
            body=content.body,
            content_type=content.content_type,
            charset=content.charset,
            attachment_filenames=None,
            attachment_binaries=None,
            attachment_binary_filenames=None,
            verbose=False
        )
        attempt = SendAttempt(
            job=job,
            config=config,
            when_attempt=now,
            success=success,
            details=details
        )
        session.add(attempt)
        if success:
            log.info("... success")
        else:
            log.error(f"... failed: {details}")
            if stop_on_error:
                log.error("Stopping because of errors.")
                return
    log.info("Done.")


# =============================================================================
# Show info
# =============================================================================

def info(session: Session) -> None:
    """
    Shows information about jobs, to stdout.
    """
    n_recipients = Recipient.n_recipients(session)
    n_content_templates = Content.n_templates(session)
    n_complete = Job.n_completed_jobs(session)
    n_pending = Job.n_pending_jobs(session)

    print(f"""Status:

Recipients: {n_recipients}
Content templates: {n_content_templates}
Jobs complete: {n_complete}
Jobs pending: {n_pending}
""")


# =============================================================================
# Clear pending jobs
# =============================================================================

def clear_pending_jobs(session: Session) -> None:
    """
    Shows information about jobs, to stdout.
    """
    log.info("Deleting pending jobs...")
    Job.clear_pending_jobs(session)
    log.info("... done.")


# =============================================================================
# Command-line entry point.
# =============================================================================

def main() -> None:
    """
    Command-line entry point.
    """

    # -------------------------------------------------------------------------
    # Arguments
    # -------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        description="Simple bulk e-mail tool. All data are stored in a "
                    "relational database (e.g. SQLite).",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Top-level arguments
    parser.add_argument(
        "--db", type=str, default=None,
        help=f"Database URL, in SQLAlchemy format. If not specified, it will "
             f"be taken from the environment variable {DB_URL_ENVVAR}"
    )
    parser.add_argument(
        "--echo", action="store_true",
        help="Echo all SQL"
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Be verbose"
    )

    # Subparsers
    # We don't use "set_defaults(func=somefunc)" because we want to have more
    # control prior to launching the subcommand.
    subparsers = parser.add_subparsers(
        help="Append '--help' to any command for further help.",
        dest="command"  # store name of chosen command here
    )
    subparsers.required = True

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: create_database_destroying_existing_data
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    _ = subparsers.add_parser(
        Command.CREATE_DB,
        help="Set up the database structure. Beware: DESTROYS EXISTING DATA.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: configure
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    parser_configure = subparsers.add_parser(
        Command.CONFIGURE,
        help="Configure the e-mailer.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_configure.add_argument(
        "--host", type=str, required=True,
        help="Mail server (SMTP) hostname."
    )
    parser_configure.add_argument(
        "--port", type=str, default=STANDARD_TLS_PORT,
        help=f"Mail server TCP port (e.g. {STANDARD_SMTP_PORT} for plain "
             f"SMTP, {STANDARD_TLS_PORT} for TLS)."
    )
    # Requires Python 3.9:
    # action=argparse.BooleanOptionalAction,
    parser_configure.add_argument(
        "--tls", action="store_true", default=True,
        help="Use TLS? (Enabled by default.)"
    )
    parser_configure.add_argument(
        "--no-tls", dest="tls", action="store_false",
        help="Disable TLS?"
    )
    parser_configure.add_argument(
        "--username", type=str, required=True,
        help="Username on mail server."
    )
    parser_configure.add_argument(
        "--password", type=str, required=True,
        help="Password on mail server. (WILL BE STORED IN DATABASE.)"
    )
    parser_configure.add_argument(
        "--time_between_emails", type=float,
        default=DEFAULT_TIME_BETWEEN_EMAILS_S,
        help="Time (in seconds) between each e-mail to be sent. Your server "
             "may object if you send them too fast."
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: add
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    parser_add_job = subparsers.add_parser(
        Command.ADD,
        help="Adds a bulk e-mail job.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_add_job.add_argument(
        "--recipients_file", type=str, required=True,
        help="Name of a file containing one recipient e-mail address per line."
    )
    parser_add_job.add_argument(
        "--from_addr", type=str, required=True,
        help="From address."
    )
    parser_add_job.add_argument(
        "--reply_to_addr", type=str,
        help="Reply-To address."
    )
    parser_add_job.add_argument(
        "--subject", type=str, required=True,
        help="E-mail subject."
    )
    contents_group = parser_add_job.add_mutually_exclusive_group(required=True)
    contents_group.add_argument(
        "--content_html_file", type=str,
        help="E-mail contents as HTML."
    )
    contents_group.add_argument(
        "--content_text_file", type=str,
        help="E-mail contents as plain text."
    )
    parser_add_job.add_argument(
        "--charset", type=str, default=UTF8,
        help="Encoding (character set)."
    )
    parser_add_job.add_argument(
        "--timestamp_at_creation", action="store_true",
        help="Set the e-mail's date/time to now, as it's created, rather than "
             "when each copy is sent (the default)."
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: work
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    parser_work = subparsers.add_parser(
        Command.WORK,
        help="Does work that is pending.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_work.add_argument(
        "--stop_on_error", action="store_true", default=True,
        help="Stop if an error occurs. (The default.)"
    )
    parser_work.add_argument(
        "--continue_on_error", dest="stop_on_error", action="store_false",
        help="Continue if an error occors."
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: info
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    _ = subparsers.add_parser(
        Command.INFO,
        help="Shows information about the jobs (queued and done).",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Subcommand: clear
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    _ = subparsers.add_parser(
        Command.CLEAR,
        help="Clear pending jobs.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Parse arguments
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    args = parser.parse_args()

    # -------------------------------------------------------------------------
    # Logging
    # -------------------------------------------------------------------------
    main_only_quicksetup_rootlogger(level=logging.DEBUG if args.verbose
                                    else logging.INFO)
    if args.verbose:
        # Show arguments, but obscuring any password.
        safe_args = deepcopy(args)
        if hasattr(safe_args, "password"):
            safe_args.password = PASSWORD_OBSCURING_STRING
        log.debug(f"Arguments: {safe_args}")

    # -------------------------------------------------------------------------
    # Database (always required)
    # -------------------------------------------------------------------------
    db_url = args.db or os.environ.get(DB_URL_ENVVAR)
    if not db_url:
        log.critical(f"Must specify the database URL, either via '--db' or "
                     f"via the environment variable {DB_URL_ENVVAR}")
        sys.exit(EXIT_FAILURE)
    engine = create_engine(db_url, echo=args.echo)
    log.info(f"Using database: {get_safe_url_from_engine(engine)}")
    session = Session(engine)

    # -------------------------------------------------------------------------
    # Launch subcommand
    # -------------------------------------------------------------------------
    command = args.command

    if command == Command.CREATE_DB:
        create_db(engine)

    elif command == Command.CONFIGURE:
        configure(
            session=session,
            host=args.host,
            port=args.port,
            use_tls=args.tls,
            username=args.username,
            password=args.password,
            time_between_emails=args.time_between_emails
        )

    elif command == Command.ADD:
        add_job(
            session=session,
            recipients_filename=args.recipients_file,
            from_addr=args.from_addr,
            reply_to_addr=args.reply_to_addr,
            subject=args.subject,
            content_html_filename=args.content_html_file,
            content_text_filename=args.content_text_file,
            charset=args.charset,
            timestamp_at_creation=args.timestamp_at_creation
        )

    elif command == Command.WORK:
        work(
            session=session,
            stop_on_error=args.stop_on_error
        )

    elif command == Command.INFO:
        info(session)

    elif command == Command.CLEAR:
        clear_pending_jobs(session)

    else:
        raise ValueError(f"Unknown command: {command}")

    # -------------------------------------------------------------------------
    # Commit
    # -------------------------------------------------------------------------
    session.commit()

    # -------------------------------------------------------------------------
    # Done
    # -------------------------------------------------------------------------
    sys.exit(EXIT_SUCCESS)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# cardinal_pythonlib/tools/convert_athena_ohdsi_codes.py

r"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

Convert SNOMED-CT codes, and their children, to OPCS4 procedure codes, given
an appropriate SNOMED-and-OPCS download from Athena OHDSI; see See
https://athena.ohdsi.org/. More generally, does this for any pair of
vocabularies.

Example:

.. code-block:: bash

    cardinalpythonlib_convert_athena_ohdsi_codes 175898006 118677009 265764009 --src_vocabulary SNOMED --descendants --dest_vocabulary OPCS4 > renal_procedures_opcs4.txt
    # ... kidney operation, procedure on urinary system, renal dialysis

"""  # noqa

import argparse
import logging
import os
from typing import Iterable, List, Set

from cardinal_pythonlib.athena_ohdsi import (
    AthenaConceptRow,
    AthenaRelationshipId,
    AthenaVocabularyId,
    get_athena_concepts,
    get_athena_concept_relationships,
)
from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger

log = logging.getLogger()

DEFAULT_CONCEPT = os.path.join(os.getcwd(), "CONCEPT.csv")
DEFAULT_CONCEPT_RELATIONSHIP = os.path.join(os.getcwd(),
                                            "CONCEPT_RELATIONSHIP.csv")


def report(concepts: Iterable[AthenaConceptRow]) -> str:
    descriptions = [str(c) for c in concepts]
    return "\n".join(sorted(descriptions))


def print_equivalent_opcs_codes(source_vocabulary: str,
                                source_codes: List[int],
                                destination_vocabulary: str,
                                concept_file: str,
                                concept_relationship_file: str,
                                with_descendants: bool = False) -> None:
    """
    Print codes from another vocabulary equivalent to the supplied source
    codes and (optionally) their descendants.

    Args:
        source_vocabulary:
            source vocabulary, e.g. ``SNOMED``
        source_codes:
            list of source (e.g. SNOMED-CT) codes to use
        destination_vocabulary:
            destination vocabulary
        concept_file:
            Athena OHDSI CONCEPT.csv TSV file, containing both vocabularies
        concept_relationship_file:
            Athena OHDSI CONCEPT_RELATIONSHIP.csv TSV file, containing both
            vocabularies
        with_descendants:
            include all descendants of the codes specified?
    """
    if not source_codes:
        log.error("No starting codes.")
        return
    log.debug(f"Starting with source codes: {source_codes!r}")
    log.debug(f"Concepts file: {concept_file}")
    log.debug(f"Concept relationship file: {concept_relationship_file}")

    equivalent_relationships = [
        AthenaRelationshipId.IS_A
    ]
    child_parent_relationships = [
        AthenaRelationshipId.MAPS_TO,
        AthenaRelationshipId.MAPPED_FROM,
        AthenaRelationshipId.SUBSUMES,
    ]
    all_relationships_of_interest = (
        equivalent_relationships +
        child_parent_relationships
    )

    # Since we are scanning many times, cache what we care about:

    concept_rows = get_athena_concepts(
        tsv_filename=concept_file,
        vocabulary_ids=[source_vocabulary, destination_vocabulary],
    )
    cr_rows = get_athena_concept_relationships(
        tsv_filename=concept_relationship_file,
        relationship_id_values=all_relationships_of_interest
    )

    # 1. Find Athena concepts from source codes
    source_codes_str = [str(x) for x in source_codes]
    parent_concepts = get_athena_concepts(
        cached_concepts=concept_rows,
        vocabulary_ids=[source_vocabulary],
        concept_codes=source_codes_str,
    )
    log.info(f"Athena concepts for starting source codes:\n"
             f"{report(parent_concepts)}")

    # 2. Find children
    source_concept_ids = set(p.concept_id for p in parent_concepts)
    if with_descendants:
        parents_to_search = source_concept_ids
        # noinspection PyTypeChecker
        ignore = set()  # type: Set[int]
        while True:
            log.debug("Iterating to find children...")
            # log.debug(f"parents_to_search = {parents_to_search}")
            # log.debug(f"ignore = {ignore}")
            new_children = [
                relrow.concept_id_1
                for relrow in get_athena_concept_relationships(
                    cached_concept_relationships=cr_rows,
                    relationship_id_values=equivalent_relationships,
                    concept_id_2_values=parents_to_search,
                    not_concept_id_1_values=ignore
                )
            ]
            if not new_children:
                break
            ignore.update(parents_to_search)  # don't search these twice
            parents_to_search = new_children
            source_concept_ids.update(new_children)
            log.debug(f"Currently have {len(source_concept_ids)} "
                      f"source concepts")
            # log.debug(f"... {source_concept_ids}")

    # Cosmetic only...
    source_concepts = get_athena_concepts(
        cached_concepts=concept_rows,
        concept_ids=source_concept_ids
    )
    log.debug(f"All source concepts:\n"
              f"{report(source_concepts)}")
    log.debug(f"source_concept_ids = {source_concept_ids}")

    # 3. Find equivalent concepts in the destination vocabulary
    destination_concept_ids = set(
        relrow.concept_id_2
        for relrow in get_athena_concept_relationships(
            cached_concept_relationships=cr_rows,
            concept_id_1_values=source_concept_ids,
            relationship_id_values=child_parent_relationships
        )
    ) - source_concept_ids
    # There are plenty of codes that are listed as mapping to themselves; we
    # ignore those (by subtracting descendant_concept_ids).
    log.debug(f"Athena concepts for equivalents: {destination_concept_ids!r}")
    if not destination_concept_ids:
        log.error("No equivalents.")
        return

    # 4. Find the actual OPCS codes
    dest_rows = sorted(
        get_athena_concepts(
            tsv_filename=concept_file,
            vocabulary_ids=[destination_vocabulary],
            concept_ids=destination_concept_ids,
        ),
        key=lambda cr: cr.concept_code
    )
    log.info(f"Destination ({destination_vocabulary}) equivalents follow.")
    for r in dest_rows:
        print(r)


def main() -> None:
    """
    Command-line entry point.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "source_codes", nargs="+", type=int,
        help="Source codes to look up "
             "(along with their descendants if --descendants is specified)"
    )
    parser.add_argument(
        "--descendants", action="store_true",
        help="Include descendants of the codes specified"
    )
    parser.add_argument(
        "--concept", type=str, default=DEFAULT_CONCEPT,
        help="Athena OHDSI CONCEPT.csv TSV file including the source and "
             "destination vocabularies"
    )
    parser.add_argument(
        "--concept_relationship", type=str,
        default=DEFAULT_CONCEPT_RELATIONSHIP,
        help="Athena OHDSI CONCEPT_RELATIONSHIP.csv TSV file "
             "including the source and destination vocabularies"
    )
    parser.add_argument(
        "--src_vocabulary", type=str, default=AthenaVocabularyId.SNOMED,
        help="Source vocabulary"
    )
    parser.add_argument(
        "--dest_vocabulary", type=str, default=AthenaVocabularyId.OPCS4,
        help="Destination vocabulary"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger()
    print_equivalent_opcs_codes(
        source_codes=args.source_codes,
        source_vocabulary=args.src_vocabulary,
        destination_vocabulary=args.dest_vocabulary,
        concept_file=args.concept,
        concept_relationship_file=args.concept_relationship,
        with_descendants=args.descendants,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# cardinal_pythonlib/process.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for process/external command management.**

"""

import shlex
import subprocess
import sys
import traceback
from typing import BinaryIO, List, Sequence, Set, Tuple

import psutil

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Get otput of extermnal commands
# =============================================================================

def get_external_command_output(command: str) -> bytes:
    """
    Takes a command-line command, executes it, and returns its ``stdout``
    output.

    Args:
        command: command string

    Returns:
        output from the command as ``bytes``

    """
    args = shlex.split(command)
    ret = subprocess.check_output(args)  # this needs Python 2.7 or higher
    return ret


def get_pipe_series_output(commands: Sequence[str],
                           stdinput: BinaryIO = None) -> bytes:
    """
    Get the output from a piped series of commands.

    Args:
        commands: sequence of command strings
        stdinput: optional ``stdin`` data to feed into the start of the pipe

    Returns:
        ``stdout`` from the end of the pipe

    """
    # Python arrays indexes are zero-based, i.e. an array is indexed from
    # 0 to len(array)-1.
    # The range/xrange commands, by default, start at 0 and go to one less
    # than the maximum specified.

    # print commands
    processes = []  # type: List[subprocess.Popen]
    for i in range(len(commands)):
        if i == 0:  # first processes
            processes.append(
                subprocess.Popen(
                    shlex.split(commands[i]),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE
                )
            )
        else:  # subsequent ones
            processes.append(
                subprocess.Popen(
                    shlex.split(commands[i]),
                    stdin=processes[i - 1].stdout,
                    stdout=subprocess.PIPE
                )
            )
    return processes[len(processes) - 1].communicate(stdinput)[0]
    # communicate() returns a tuple; 0=stdout, 1=stderr; so this returns stdout

# Also, simple commands: use os.system(command)


# =============================================================================
# Launch external file using OS's launcher
# =============================================================================

def launch_external_file(filename: str, raise_if_fails: bool = False) -> None:
    """
    Launches a file using the operating system's standard launcher.

    Args:
        filename: file to launch
        raise_if_fails: raise any exceptions from
            ``subprocess.call(["xdg-open", filename])`` (Linux)
            or ``os.startfile(filename)`` (otherwise)? If not, exceptions
            are suppressed.

    """
    log.info("Launching external file: {!r}", filename)
    try:
        if sys.platform.startswith('linux'):
            cmdargs = ["xdg-open", filename]
            # log.debug("... command: {!r}", cmdargs)
            subprocess.call(cmdargs)
        else:
            # log.debug("... with os.startfile()")
            # noinspection PyUnresolvedReferences
            os.startfile(filename)
    except Exception as e:
        log.critical("Error launching {!r}: error was {}.\n\n{}",
                     filename, str(e), traceback.format_exc())
        if raise_if_fails:
            raise


# =============================================================================
# Kill a process tree. Particularly useful for Windows, where a plain "kill()"
# (via subprocess) can leave orphans.
# =============================================================================

def kill_proc_tree(pid: int,
                   including_parent: bool = True,
                   timeout_s: float = 5) \
        -> Tuple[Set[psutil.Process], Set[psutil.Process]]:
    """
    Kills a tree of processes, starting with the parent. Slightly modified from
    https://stackoverflow.com/questions/1230669/subprocess-deleting-child-processes-in-windows.

    Args:
        pid: process ID of the parent
        including_parent: kill the parent too?
        timeout_s: timeout to wait for processes to close

    Returns:
        tuple: ``(gone, still_alive)``, where both are sets of
        :class:`psutil.Process` objects

    """  # noqa
    parent = psutil.Process(pid)
    to_kill = parent.children(recursive=True)  # type: List[psutil.Process]
    if including_parent:
        to_kill.append(parent)
    for proc in to_kill:
        proc.kill()  # SIGKILL
    gone, still_alive = psutil.wait_procs(to_kill, timeout=timeout_s)
    return gone, still_alive


# =============================================================================
# nice_call
# =============================================================================

def nice_call(*popenargs, timeout: float = None,
              cleanup_timeout: float = None, **kwargs) -> int:
    """
    Like :func:`subprocess.call`, but give the child process time to
    clean up and communicate if a :exc:`KeyboardInterrupt` is raised.

    Modified from
    https://stackoverflow.com/questions/34458583/python-subprocess-call-doesnt-handle-signal-correctly
    """  # noqa
    with subprocess.Popen(*popenargs, **kwargs) as p:
        try:
            return p.wait(timeout=timeout)
        except KeyboardInterrupt:
            log.error("KeyboardInterrupt received")
            if cleanup_timeout:
                # Wait again, now that the child has received SIGINT, too.
                log.info(f"Waiting {cleanup_timeout} seconds "
                         f"for child process {p.pid} to finish...")
                try:
                    p.wait(timeout=cleanup_timeout)  # may raise TimeoutExpired
                    log.info(f"Child process {p.pid} shut down cleanly")
                except subprocess.TimeoutExpired:
                    log.info(f"Killing child process {p.pid}")
                    p.kill()
                    p.wait()
            raise  # propagate KeyboardInterrupt up through Python program
        except Exception:
            log.error(f"Error with child process {p.pid}. Killing it...")
            p.kill()
            p.wait()
            log.info(f"Child process {p.pid} killed")
            raise



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/orm_query.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to perform and manipulate SQLAlchemy ORM queries.**

"""

from typing import Any, Dict, Sequence, Tuple, Union

from sqlalchemy.engine.base import Connection, Engine
from sqlalchemy.engine.result import ResultProxy
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.expression import ClauseElement, literal
from sqlalchemy.sql import func
from sqlalchemy.sql.selectable import Exists

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.sqlalchemy.dialect import SqlaDialectName

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Get query result with fieldnames
# =============================================================================

def get_rows_fieldnames_from_query(
        session: Union[Session, Engine, Connection],
        query: Query) -> Tuple[Sequence[Sequence[Any]], Sequence[str]]:
    """
    Returns results and column names from a query.

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        query: SQLAlchemy :class:`Query`

    Returns:
        ``(rows, fieldnames)`` where ``rows`` is the usual set of results and
        ``fieldnames`` are the name of the result columns/fields.

    """
    # https://stackoverflow.com/questions/6455560/how-to-get-column-names-from-sqlalchemy-result-declarative-syntax  # noqa
    # No! Returns e.g. "User" for session.Query(User)...
    # fieldnames = [cd['name'] for cd in query.column_descriptions]
    result = session.execute(query)  # type: ResultProxy
    fieldnames = result.keys()
    # ... yes! Comes out as "_table_field", which is how SQLAlchemy SELECTs
    # things.
    rows = result.fetchall()
    return rows, fieldnames


# =============================================================================
# EXISTS (SQLAlchemy ORM)
# =============================================================================

def bool_from_exists_clause(session: Session,
                            exists_clause: Exists) -> bool:
    """
    Database dialects are not consistent in how ``EXISTS`` clauses can be
    converted to a boolean answer. This function manages the inconsistencies.

    See:

    - https://bitbucket.org/zzzeek/sqlalchemy/issues/3212/misleading-documentation-for-queryexists
    - https://docs.sqlalchemy.org/en/latest/orm/query.html#sqlalchemy.orm.query.Query.exists

    Specifically, we want this:

    *SQL Server*

    .. code-block:: sql

        SELECT 1 WHERE EXISTS (SELECT 1 FROM table WHERE ...)
        -- ... giving 1 or None (no rows)
        -- ... fine for SQL Server, but invalid for MySQL (no FROM clause)

    *Others, including MySQL*

    .. code-block:: sql

        SELECT EXISTS (SELECT 1 FROM table WHERE ...)
        -- ... giving 1 or 0
        -- ... fine for MySQL, but invalid syntax for SQL Server

    """  # noqa
    if session.get_bind().dialect.name == SqlaDialectName.MSSQL:
        # SQL Server
        result = session.query(literal(True)).filter(exists_clause).scalar()
    else:
        # MySQL, etc.
        result = session.query(exists_clause).scalar()
    return bool(result)


def exists_orm(session: Session,
               ormclass: DeclarativeMeta,
               *criteria: Any) -> bool:
    """
    Detects whether a database record exists for the specified ``ormclass``
    and ``criteria``.

    Example usage:

    .. code-block:: python

        bool_exists = exists_orm(session, MyClass, MyClass.myfield == value)
    """
    # http://docs.sqlalchemy.org/en/latest/orm/query.html
    q = session.query(ormclass)
    for criterion in criteria:
        q = q.filter(criterion)
    exists_clause = q.exists()
    return bool_from_exists_clause(session=session,
                                   exists_clause=exists_clause)


# =============================================================================
# Get or create (SQLAlchemy ORM)
# =============================================================================

def get_or_create(session: Session,
                  model: DeclarativeMeta,
                  defaults: Dict[str, Any] = None,
                  **kwargs: Any) -> Tuple[Any, bool]:
    """
    Fetches an ORM object from the database, or creates one if none existed.

    Args:
        session: an SQLAlchemy :class:`Session`
        model: an SQLAlchemy ORM class
        defaults: default initialization arguments (in addition to relevant
            filter criteria) if we have to create a new instance
        kwargs: optional filter criteria

    Returns:
        a tuple ``(instance, newly_created)``

    See https://stackoverflow.com/questions/2546207 (this function is a
    composite of several suggestions).
    """
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance, False
    else:
        params = dict((k, v) for k, v in kwargs.items()
                      if not isinstance(v, ClauseElement))
        params.update(defaults or {})
        instance = model(**params)
        session.add(instance)
        return instance, True


# =============================================================================
# Extend Query to provide an optimized COUNT(*)
# =============================================================================

# noinspection PyAbstractClass
class CountStarSpecializedQuery(Query):
    def __init__(self, *args, **kwargs) -> None:
        """
        Optimizes ``COUNT(*)`` queries.

        See
        https://stackoverflow.com/questions/12941416/how-to-count-rows-with-select-count-with-sqlalchemy

        Example use:

        .. code-block:: python

            q = CountStarSpecializedQuery([cls], session=dbsession)\
                .filter(cls.username == username)
            return q.count_star()

        """  # noqa
        super().__init__(*args, **kwargs)

    def count_star(self) -> int:
        """
        Implements the ``COUNT(*)`` specialization.
        """
        count_query = (self.statement.with_only_columns([func.count()])
                       .order_by(None))
        return self.session.execute(count_query).scalar()



#!/usr/bin/env python
# cardinal_pythonlib/sphinxtools.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to help with Sphinx, in particular the generation of autodoc
files.**

Rationale: if you want Sphinx ``autodoc`` code to appear as "one module per
Sphinx page" (which I normally do), you need one ``.rst`` file per module.

"""

from enum import Enum
from fnmatch import fnmatch
import glob
import logging
from os.path import (
    abspath, basename, dirname, exists, expanduser, isdir, isfile, join,
    relpath, sep, splitext
)
from typing import Dict, Iterable, List, NoReturn, Union

from cardinal_pythonlib.fileops import mkdir_p, relative_filename_within_dir
from cardinal_pythonlib.logs import BraceStyleAdapter
from cardinal_pythonlib.reprfunc import auto_repr
from pygments.lexer import Lexer
from pygments.lexers import get_lexer_for_filename
from pygments.util import ClassNotFound

log = BraceStyleAdapter(logging.getLogger(__name__))


# =============================================================================
# Constants
# =============================================================================

AUTOGENERATED_COMMENT = ".. THIS FILE IS AUTOMATICALLY GENERATED. DO NOT EDIT."
DEFAULT_INDEX_TITLE = "Automatic documentation of source code"
DEFAULT_SKIP_GLOBS = ["__init__.py"]
EXT_PYTHON = ".py"
EXT_RST = ".rst"
CODE_TYPE_NONE = "none"


class AutodocMethod(Enum):
    """
    Enum to specify the method of autodocumenting a file.
    """
    BEST = 0
    CONTENTS = 1
    AUTOMODULE = 2


# =============================================================================
# Helper functions
# =============================================================================

def rst_underline(heading: str, underline_char: str) -> str:
    """
    Underlines a heading for RST files.

    Args:
        heading: text to underline
        underline_char: character to use

    Returns:
        underlined heading, over two lines (without a final terminating
        newline)
    """
    assert "\n" not in heading
    assert len(underline_char) == 1
    return heading + "\n" + (underline_char * len(heading))


def fail(msg: str) -> NoReturn:
    log.critical(msg)
    raise RuntimeError(msg)


def write_if_allowed(filename: str,
                     content: str,
                     overwrite: bool = False,
                     mock: bool = False) -> None:
    """
    Writes the contents to a file, if permitted.

    Args:
        filename: filename to write
        content: contents to write
        overwrite: permit overwrites?
        mock: pretend to write, but don't

    Raises:
        RuntimeError: if file exists but overwriting not permitted
    """
    # Check we're allowed
    if not overwrite and exists(filename):
        fail(f"File exists, not overwriting: {filename!r}")

    # Make the directory, if necessary
    directory = dirname(filename)
    if not mock:
        mkdir_p(directory)

    # Write the file
    log.info("Writing to {!r}", filename)
    if mock:
        log.warning("Skipping writes as in mock mode")
    else:
        with open(filename, "wt") as outfile:
            outfile.write(content)


# =============================================================================
# FileToAutodocument
# =============================================================================

class FileToAutodocument(object):
    """
    Class representing a file to document automatically via Sphinx autodoc.

    Example:
        
    .. code-block:: python

        import logging
        from cardinal_pythonlib.logs import *
        from cardinal_pythonlib.sphinxtools import *
        main_only_quicksetup_rootlogger(level=logging.DEBUG)
        
        f = FileToAutodocument(
            source_filename="~/Documents/code/cardinal_pythonlib/cardinal_pythonlib/sphinxtools.py",
            project_root_dir="~/Documents/code/cardinal_pythonlib",
            target_rst_filename="~/Documents/code/cardinal_pythonlib/docs/source/autodoc/sphinxtools.rst",
        )
        print(f)
        f.source_extension
        f.is_python
        f.source_filename_rel_project_root
        f.rst_dir
        f.source_filename_rel_rst_file
        f.rst_filename_rel_project_root
        f.rst_filename_rel_autodoc_index(
            "~/Documents/code/cardinal_pythonlib/docs/source/autodoc/_index.rst")
        f.python_module_name
        f.pygments_code_type
        print(f.rst_content(prefix=".. Hello!"))
        print(f.rst_content(prefix=".. Hello!", method=AutodocMethod.CONTENTS))
        f.write_rst(prefix=".. Hello!")

    """  # noqa

    def __init__(self,
                 source_filename: str,
                 project_root_dir: str,
                 target_rst_filename: str,
                 method: AutodocMethod = AutodocMethod.BEST,
                 python_package_root_dir: str = None,
                 source_rst_title_style_python: bool = True,
                 pygments_language_override: Dict[str, str] = None) -> None:
        """
        Args:
            source_filename: source file (e.g. Python, C++, XML file) to
                document
            project_root_dir: root directory of the whole project
            target_rst_filename: filenamd of an RST file to write that will
                document the source file
            method: instance of :class:`AutodocMethod`; for example, should we
                ask Sphinx's ``autodoc`` to read docstrings and build us a
                pretty page, or just include the contents with syntax
                highlighting?
            python_package_root_dir: if your Python modules live in a directory
                other than ``project_root_dir``, specify it here
            source_rst_title_style_python: if ``True`` and the file is a Python
                file and ``method == AutodocMethod.AUTOMODULE``, the heading
                used will be in the style of a Python module, ``x.y.z``.
                Otherwise, it will be a path (``x/y/z``).
            pygments_language_override: if specified, a dictionary mapping
                file extensions to Pygments languages (for example: a ``.pro``
                file will be autodetected as Prolog, but you might want to
                map that to ``none`` for Qt project files).
        """
        self.source_filename = abspath(expanduser(source_filename))
        self.project_root_dir = abspath(expanduser(project_root_dir))
        self.target_rst_filename = abspath(expanduser(target_rst_filename))
        self.method = method
        self.source_rst_title_style_python = source_rst_title_style_python
        self.python_package_root_dir = (
            abspath(expanduser(python_package_root_dir))
            if python_package_root_dir else self.project_root_dir
        )
        self.pygments_language_override = pygments_language_override or {}  # type: Dict[str, str]  # noqa
        assert isfile(self.source_filename), (
            f"Not a file: source_filename={self.source_filename!r}")
        assert isdir(self.project_root_dir), (
            f"Not a directory: project_root_dir={self.project_root_dir!r}")
        assert relative_filename_within_dir(
            filename=self.source_filename,
            directory=self.project_root_dir
        ), (
            f"Source file {self.source_filename!r} is not within "
            f"project directory {self.project_root_dir!r}"
        )
        assert relative_filename_within_dir(
            filename=self.python_package_root_dir,
            directory=self.project_root_dir
        ), (
            f"Python root {self.python_package_root_dir!r} is not within "
            f"project directory {self.project_root_dir!r}"
        )
        assert isinstance(method, AutodocMethod)

    def __repr__(self) -> str:
        return auto_repr(self)

    @property
    def source_extension(self) -> str:
        """
        Returns the extension of the source filename.
        """
        return splitext(self.source_filename)[1]

    @property
    def is_python(self) -> bool:
        """
        Is the source file a Python file?
        """
        return self.source_extension == EXT_PYTHON

    @property
    def source_filename_rel_project_root(self) -> str:
        """
        Returns the name of the source filename, relative to the project root.
        Used to calculate file titles.
        """
        return relpath(self.source_filename, start=self.project_root_dir)

    @property
    def source_filename_rel_python_root(self) -> str:
        """
        Returns the name of the source filename, relative to the Python package
        root. Used to calculate the name of Python modules.
        """
        return relpath(self.source_filename,
                       start=self.python_package_root_dir)

    @property
    def rst_dir(self) -> str:
        """
        Returns the directory of the target RST file.
        """
        return dirname(self.target_rst_filename)

    @property
    def source_filename_rel_rst_file(self) -> str:
        """
        Returns the source filename as seen from the RST filename that we
        will generate. Used for ``.. include::`` commands.
        """
        return relpath(self.source_filename, start=self.rst_dir)

    @property
    def rst_filename_rel_project_root(self) -> str:
        """
        Returns the filename of the target RST file, relative to the project
        root directory. Used for labelling the RST file itself.
        """
        return relpath(self.target_rst_filename, start=self.project_root_dir)

    def rst_filename_rel_autodoc_index(self, index_filename: str) -> str:
        """
        Returns the filename of the target RST file, relative to a specified
        index file. Used to make the index refer to the RST.
        """
        index_dir = dirname(abspath(expanduser(index_filename)))
        return relpath(self.target_rst_filename, start=index_dir)

    @property
    def python_module_name(self) -> str:
        """
        Returns the name of the Python module that this instance refers to,
        in dotted Python module notation, or a blank string if it doesn't.
        """
        if not self.is_python:
            return ""
        filepath = self.source_filename_rel_python_root
        dirs_and_base = splitext(filepath)[0]
        dir_and_file_parts = dirs_and_base.split(sep)
        return ".".join(dir_and_file_parts)

    @property
    def pygments_language(self) -> str:
        """
        Returns the code type annotation for Pygments; e.g. ``python`` for
        Python, ``cpp`` for C++, etc.
        """
        extension = splitext(self.source_filename)[1]
        if extension in self.pygments_language_override:
            return self.pygments_language_override[extension]
        try:
            lexer = get_lexer_for_filename(self.source_filename)  # type: Lexer
            return lexer.name
        except ClassNotFound:
            log.warning("Don't know Pygments code type for extension {!r}",
                        self.source_extension)
            return CODE_TYPE_NONE

    def rst_content(self,
                    prefix: str = "",
                    suffix: str = "",
                    heading_underline_char: str = "=",
                    method: AutodocMethod = None) -> str:
        """
        Returns the text contents of an RST file that will automatically
        document our source file.

        Args:
            prefix: prefix, e.g. RST copyright comment
            suffix: suffix, after the part we're creating
            heading_underline_char: RST character to use to underline the
                heading
            method: optional method to override ``self.method``; see
                constructor

        Returns:
            the RST contents
        """
        spacer = "    "
        # Choose our final method
        if method is None:
            method = self.method
        is_python = self.is_python
        if method == AutodocMethod.BEST:
            if is_python:
                method = AutodocMethod.AUTOMODULE
            else:
                method = AutodocMethod.CONTENTS
        elif method == AutodocMethod.AUTOMODULE:
            if not is_python:
                method = AutodocMethod.CONTENTS

        # Write the instruction
        if method == AutodocMethod.AUTOMODULE:
            if self.source_rst_title_style_python:
                title = self.python_module_name
            else:
                title = self.source_filename_rel_project_root
            instruction = (
                f".. automodule:: {self.python_module_name}\n"
                f"    :members:"
            )
        elif method == AutodocMethod.CONTENTS:
            title = self.source_filename_rel_project_root
            # Using ".. include::" with options like ":code: python" doesn't
            # work properly; everything comes out as Python.
            # Instead, see http://www.sphinx-doc.org/en/1.4.9/markup/code.html;
            # we need ".. literalinclude::" with ":language: LANGUAGE".

            instruction = (
                ".. literalinclude:: {filename}\n"
                "{spacer}:language: {language}".format(
                    filename=self.source_filename_rel_rst_file,
                    spacer=spacer,
                    language=self.pygments_language
                )
            )
        else:
            raise ValueError("Bad method!")

        # Create the whole file
        content = """
.. {filename}

{AUTOGENERATED_COMMENT}

{prefix}

{underlined_title}

{instruction}

{suffix}
        """.format(
            filename=self.rst_filename_rel_project_root,
            AUTOGENERATED_COMMENT=AUTOGENERATED_COMMENT,
            prefix=prefix,
            underlined_title=rst_underline(
                title, underline_char=heading_underline_char),
            instruction=instruction,
            suffix=suffix,
        ).strip() + "\n"
        return content

    def write_rst(self,
                  prefix: str = "",
                  suffix: str = "",
                  heading_underline_char: str = "=",
                  method: AutodocMethod = None,
                  overwrite: bool = False,
                  mock: bool = False) -> None:
        """
        Writes the RST file to our destination RST filename, making any
        necessary directories.

        Args:
            prefix: as for :func:`rst_content`
            suffix: as for :func:`rst_content`
            heading_underline_char: as for :func:`rst_content`
            method: as for :func:`rst_content`
            overwrite: overwrite the file if it exists already?
            mock: pretend to write, but don't
        """
        content = self.rst_content(
            prefix=prefix,
            suffix=suffix,
            heading_underline_char=heading_underline_char,
            method=method
        )
        write_if_allowed(self.target_rst_filename, content,
                         overwrite=overwrite, mock=mock)


# =============================================================================
# AutodocIndex
# =============================================================================

class AutodocIndex(object):
    """
    Class to make an RST file that indexes others.

    Example:

    .. code-block:: python

        import logging
        from cardinal_pythonlib.logs import *
        from cardinal_pythonlib.sphinxtools import *
        main_only_quicksetup_rootlogger(level=logging.INFO)
        
        # Example where one index contains another:
        
        subidx = AutodocIndex(
            index_filename="~/Documents/code/cardinal_pythonlib/docs/source/autodoc/_index2.rst",
            highest_code_dir="~/Documents/code/cardinal_pythonlib",
            project_root_dir="~/Documents/code/cardinal_pythonlib",
            autodoc_rst_root_dir="~/Documents/code/cardinal_pythonlib/docs/source/autodoc",
            source_filenames_or_globs="~/Documents/code/cardinal_pythonlib/docs/*.py",
        )
        idx = AutodocIndex(
            index_filename="~/Documents/code/cardinal_pythonlib/docs/source/autodoc/_index.rst",
            highest_code_dir="~/Documents/code/cardinal_pythonlib",
            project_root_dir="~/Documents/code/cardinal_pythonlib",
            autodoc_rst_root_dir="~/Documents/code/cardinal_pythonlib/docs/source/autodoc",
            source_filenames_or_globs="~/Documents/code/cardinal_pythonlib/cardinal_pythonlib/*.py",
        )
        idx.add_index(subidx)
        print(idx.index_content())
        idx.write_index_and_rst_files(overwrite=True, mock=True)
        
        # Example with a flat index:
        
        flatidx = AutodocIndex(
            index_filename="~/Documents/code/cardinal_pythonlib/docs/source/autodoc/_index.rst",
            highest_code_dir="~/Documents/code/cardinal_pythonlib/cardinal_pythonlib",
            project_root_dir="~/Documents/code/cardinal_pythonlib",
            autodoc_rst_root_dir="~/Documents/code/cardinal_pythonlib/docs/source/autodoc",
            source_filenames_or_globs="~/Documents/code/cardinal_pythonlib/cardinal_pythonlib/*.py",
        )
        print(flatidx.index_content())
        flatidx.write_index_and_rst_files(overwrite=True, mock=True)

    """  # noqa
    def __init__(self,
                 index_filename: str,
                 project_root_dir: str,
                 autodoc_rst_root_dir: str,
                 highest_code_dir: str,
                 python_package_root_dir: str = None,
                 source_filenames_or_globs: Union[str, Iterable[str]] = None,
                 index_heading_underline_char: str = "-",
                 source_rst_heading_underline_char: str = "~",
                 title: str = DEFAULT_INDEX_TITLE,
                 introductory_rst: str = "",
                 recursive: bool = True,
                 skip_globs: List[str] = None,
                 toctree_maxdepth: int = 1,
                 method: AutodocMethod = AutodocMethod.BEST,
                 rst_prefix: str = "",
                 rst_suffix: str = "",
                 source_rst_title_style_python: bool = True,
                 pygments_language_override: Dict[str, str] = None) -> None:
        """
        Args:
            index_filename:
                filename of the index ``.RST`` (ReStructured Text) file to
                create

            project_root_dir:
                top-level directory for the whole project

            autodoc_rst_root_dir:
                directory within which all automatically generated ``.RST``
                files (each to document a specific source file) will be placed.
                A directory hierarchy within this directory will be created,
                reflecting the structure of the code relative to
                ``highest_code_dir`` (q.v.).

            highest_code_dir:
                the "lowest" directory such that all code is found within it;
                the directory structure within ``autodoc_rst_root_dir`` is to
                ``.RST`` files what the directory structure is of the source
                files, relative to ``highest_code_dir``.

            python_package_root_dir:
                if your Python modules live in a directory other than
                ``project_root_dir``, specify it here

            source_filenames_or_globs:
                optional string, or list of strings, each describing a file or
                glob-style file specification; these are the source filenames
                to create automatic RST` for. If you don't specify them here,
                you can use :func:`add_source_files`. To add sub-indexes, use
                :func:`add_index` and :func:`add_indexes`.

            index_heading_underline_char:
                the character used to underline the title in the index file

            source_rst_heading_underline_char:
                the character used to underline the heading in each of the
                source files

            title:
                title for the index

            introductory_rst:
                extra RST for the index, which goes between the title and the
                table of contents

            recursive:
                use :func:`glob.glob` in recursive mode?

            skip_globs:
                list of file names or file specifications to skip; e.g.
                ``['__init__.py']``

            toctree_maxdepth:
                ``maxdepth`` parameter for the ``toctree`` command generated in
                the index file

            method:
                see :class:`FileToAutodocument`

            rst_prefix:
                optional RST content (e.g. copyright comment) to put early on
                in each of the RST files

            rst_suffix:
                optional RST content to put late on in each of the RST files

            source_rst_title_style_python:
                make the individual RST files use titles in the style of Python
                modules, ``x.y.z``, rather than path style (``x/y/z``); path
                style will be used for non-Python files in any case.

            pygments_language_override:
                if specified, a dictionary mapping file extensions to Pygments
                languages (for example: a ``.pro`` file will be autodetected as
                Prolog, but you might want to map that to ``none`` for Qt
                project files).

        """
        assert index_filename
        assert project_root_dir
        assert autodoc_rst_root_dir
        assert isinstance(toctree_maxdepth, int)
        assert isinstance(method, AutodocMethod)

        self.index_filename = abspath(expanduser(index_filename))
        self.title = title
        self.introductory_rst = introductory_rst
        self.project_root_dir = abspath(expanduser(project_root_dir))
        self.autodoc_rst_root_dir = abspath(expanduser(autodoc_rst_root_dir))
        self.highest_code_dir = abspath(expanduser(highest_code_dir))
        self.python_package_root_dir = (
            abspath(expanduser(python_package_root_dir))
            if python_package_root_dir else self.project_root_dir
        )
        self.index_heading_underline_char = index_heading_underline_char
        self.source_rst_heading_underline_char = source_rst_heading_underline_char  # noqa
        self.recursive = recursive
        self.skip_globs = skip_globs if skip_globs is not None else DEFAULT_SKIP_GLOBS  # noqa
        self.toctree_maxdepth = toctree_maxdepth
        self.method = method
        self.rst_prefix = rst_prefix
        self.rst_suffix = rst_suffix
        self.source_rst_title_style_python = source_rst_title_style_python
        self.pygments_language_override = pygments_language_override or {}  # type: Dict[str, str]  # noqa

        assert isdir(self.project_root_dir), (
            f"Not a directory: project_root_dir={self.project_root_dir!r}")
        assert relative_filename_within_dir(
            filename=self.index_filename,
            directory=self.project_root_dir
        ), (
            f"Index file {self.index_filename!r} is not within "
            f"project directory {self.project_root_dir!r}"
        )
        assert relative_filename_within_dir(
            filename=self.highest_code_dir,
            directory=self.project_root_dir
        ), (
            f"Highest code directory {self.highest_code_dir!r} is not within "
            f"project directory {self.project_root_dir!r}"
        )
        assert relative_filename_within_dir(
            filename=self.autodoc_rst_root_dir,
            directory=self.project_root_dir
        ), (
            f"Autodoc RST root directory {self.autodoc_rst_root_dir!r} is not "
            f"within project directory {self.project_root_dir!r}"
        )
        assert isinstance(method, AutodocMethod)
        assert isinstance(recursive, bool)

        self.files_to_index = []  # type: List[Union[FileToAutodocument, AutodocIndex]]  # noqa
        if source_filenames_or_globs:
            self.add_source_files(source_filenames_or_globs)

    def __repr__(self) -> str:
        return auto_repr(self)

    def add_source_files(
            self,
            source_filenames_or_globs: Union[str, List[str]],
            method: AutodocMethod = None,
            recursive: bool = None,
            source_rst_title_style_python: bool = None,
            pygments_language_override: Dict[str, str] = None) -> None:
        """
        Adds source files to the index.

        Args:
            source_filenames_or_globs: string containing a filename or a
                glob, describing the file(s) to be added, or a list of such
                strings
            method: optional method to override ``self.method``
            recursive: use :func:`glob.glob` in recursive mode? (If ``None``,
                the default, uses the version from the constructor.)
            source_rst_title_style_python: optional to override
                ``self.source_rst_title_style_python``
            pygments_language_override: optional to override
                ``self.pygments_language_override``
        """
        if not source_filenames_or_globs:
            return

        if method is None:
            # Use the default
            method = self.method
        if recursive is None:
            recursive = self.recursive
        if source_rst_title_style_python is None:
            source_rst_title_style_python = self.source_rst_title_style_python
        if pygments_language_override is None:
            pygments_language_override = self.pygments_language_override

        # Get a sorted list of filenames
        final_filenames = self.get_sorted_source_files(
            source_filenames_or_globs,
            recursive=recursive
        )

        # Process that sorted list
        for source_filename in final_filenames:
            self.files_to_index.append(FileToAutodocument(
                source_filename=source_filename,
                project_root_dir=self.project_root_dir,
                python_package_root_dir=self.python_package_root_dir,
                target_rst_filename=self.specific_file_rst_filename(
                    source_filename
                ),
                method=method,
                source_rst_title_style_python=source_rst_title_style_python,
                pygments_language_override=pygments_language_override,
            ))

    def get_sorted_source_files(
            self,
            source_filenames_or_globs: Union[str, List[str]],
            recursive: bool = True) -> List[str]:
        """
        Returns a sorted list of filenames to process, from a filename,
        a glob string, or a list of filenames/globs.

        Args:
            source_filenames_or_globs: filename/glob, or list of them
            recursive: use :func:`glob.glob` in recursive mode?

        Returns:
            sorted list of files to process
        """
        if isinstance(source_filenames_or_globs, str):
            source_filenames_or_globs = [source_filenames_or_globs]
        final_filenames = []  # type: List[str]
        for sfg in source_filenames_or_globs:
            sfg_expanded = expanduser(sfg)
            log.debug("Looking for: {!r}", sfg_expanded)
            for filename in glob.glob(sfg_expanded, recursive=recursive):
                log.debug("Trying: {!r}", filename)
                if self.should_exclude(filename):
                    log.info("Skipping file {!r}", filename)
                    continue
                final_filenames.append(filename)
        final_filenames.sort()
        return final_filenames

    @staticmethod
    def filename_matches_glob(filename: str, globtext: str) -> bool:
        """
        The ``glob.glob`` function doesn't do exclusion very well. We don't
        want to have to specify root directories for exclusion patterns. We
        don't want to have to trawl a massive set of files to find exclusion
        files. So let's implement a glob match.

        Args:
            filename: filename
            globtext: glob

        Returns:
            does the filename match the glob?

        See also:

        - https://stackoverflow.com/questions/20638040/glob-exclude-pattern

        """
        # Quick check on basename-only matching
        if fnmatch(filename, globtext):
            log.debug("{!r} matches {!r}", filename, globtext)
            return True
        bname = basename(filename)
        if fnmatch(bname, globtext):
            log.debug("{!r} matches {!r}", bname, globtext)
            return True
        # Directory matching: is actually accomplished by the code above!
        # Otherwise:
        return False

    def should_exclude(self, filename) -> bool:
        """
        Should we exclude this file from consideration?
        """
        for skip_glob in self.skip_globs:
            if self.filename_matches_glob(filename, skip_glob):
                return True
        return False

    def add_index(self, index: "AutodocIndex") -> None:
        """
        Add a sub-index file to this index.

        Args:
            index: index file to add, as an instance of :class:`AutodocIndex`
        """
        self.files_to_index.append(index)

    def add_indexes(self, indexes: List["AutodocIndex"]) -> None:
        """
        Adds multiple sub-indexes to this index.

        Args:
            indexes: list of sub-indexes
        """
        for index in indexes:
            self.add_index(index)

    def specific_file_rst_filename(self, source_filename: str) -> str:
        """
        Gets the RST filename corresponding to a source filename.
        See the help for the constructor for more details.

        Args:
            source_filename: source filename within current project

        Returns:
            RST filename

        Note in particular: the way we structure the directories means that we
        won't get clashes between files with idential names in two different
        directories. However, we must also incorporate the original source
        filename, in particular for C++ where ``thing.h`` and ``thing.cpp``
        must not generate the same RST filename. So we just add ``.rst``.
        """
        highest_code_to_target = relative_filename_within_dir(
            source_filename, self.highest_code_dir)
        bname = basename(source_filename)
        result = join(self.autodoc_rst_root_dir,
                      dirname(highest_code_to_target),
                      bname + EXT_RST)
        log.debug("Source {!r} -> RST {!r}", source_filename, result)
        return result

    def write_index_and_rst_files(self, overwrite: bool = False,
                                  mock: bool = False) -> None:
        """
        Writes both the individual RST files and the index.

        Args:
            overwrite: allow existing files to be overwritten?
            mock: pretend to write, but don't
        """
        for f in self.files_to_index:
            if isinstance(f, FileToAutodocument):
                f.write_rst(
                    prefix=self.rst_prefix,
                    suffix=self.rst_suffix,
                    heading_underline_char=self.source_rst_heading_underline_char,  # noqa
                    overwrite=overwrite,
                    mock=mock,
                )
            elif isinstance(f, AutodocIndex):
                f.write_index_and_rst_files(overwrite=overwrite, mock=mock)
            else:
                fail(f"Unknown thing in files_to_index: {f!r}")
        self.write_index(overwrite=overwrite, mock=mock)

    @property
    def index_filename_rel_project_root(self) -> str:
        """
        Returns the name of the index filename, relative to the project root.
        Used for labelling the index file.
        """
        return relpath(self.index_filename, start=self.project_root_dir)

    def index_filename_rel_other_index(self, other: str) -> str:
        """
        Returns the filename of this index, relative to the director of another
        index. (For inserting a reference to this index into ``other``.)

        Args:
            other: the other index

        Returns:
            relative filename of our index
        """
        return relpath(self.index_filename, start=dirname(other))

    def index_content(self) -> str:
        """
        Returns the contents of the index RST file.
        """
        # Build the toctree command
        index_filename = self.index_filename
        spacer = "    "
        toctree_lines = [
            "..  toctree::",
            spacer + f":maxdepth: {self.toctree_maxdepth}",
            ""
        ]
        for f in self.files_to_index:
            if isinstance(f, FileToAutodocument):
                rst_filename = spacer + f.rst_filename_rel_autodoc_index(
                    index_filename)
            elif isinstance(f, AutodocIndex):
                rst_filename = (
                    spacer + f.index_filename_rel_other_index(index_filename)
                )
            else:
                fail(f"Unknown thing in files_to_index: {f!r}")
                rst_filename = ""  # won't get here; for the type checker
            toctree_lines.append(rst_filename)
        toctree = "\n".join(toctree_lines)

        # Create the whole file
        content = """
.. {filename}

{AUTOGENERATED_COMMENT}

{prefix}

{underlined_title}

{introductory_rst}

{toctree}

{suffix}
                """.format(
            filename=self.index_filename_rel_project_root,
            AUTOGENERATED_COMMENT=AUTOGENERATED_COMMENT,
            prefix=self.rst_prefix,
            underlined_title=rst_underline(
                self.title, underline_char=self.index_heading_underline_char),
            introductory_rst=self.introductory_rst,
            toctree=toctree,
            suffix=self.rst_suffix,
        ).strip() + "\n"
        return content

    def write_index(self, overwrite: bool = False, mock: bool = False) -> None:
        """
        Writes the index file, if permitted.

        Args:
            overwrite: allow existing files to be overwritten?
            mock: pretend to write, but don't
        """
        write_if_allowed(self.index_filename, self.index_content(),
                         overwrite=overwrite, mock=mock)



#!/usr/bin/env python
# cardinal_pythonlib/contexts.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Context manager assistance.**

"""

import contextlib


@contextlib.contextmanager
def dummy_context_mgr():
    """
    We might be using Python 3.6 which doesn't have ``contextlib.nullcontext``.
    Hence this.

    - https://stackoverflow.com/questions/27803059/conditional-with-statement-in-python
    - See also
      https://stackoverflow.com/questions/893333/multiple-variables-in-a-with-statement
    """  # noqa
    yield None



#!/usr/bin/env python
# cardinal_pythonlib/email/sendmail.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Sends e-mails from the command line.**

"""

import argparse
import email.encoders
import email.mime.base
import email.mime.text
import email.mime.multipart
import email.header
import email.utils
import logging
import os
import re
import smtplib
import sys
from typing import List, NoReturn, Sequence, Tuple, Union
import unittest

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Constants
# =============================================================================

CONTENT_TYPE_TEXT = "text/plain"
CONTENT_TYPE_HTML = "text/html"

ASCII = "ascii"
UTF8 = "utf8"

COMMA = ","
COMMASPACE = ", "

STANDARD_SMTP_PORT = 25
STANDARD_TLS_PORT = 587


# =============================================================================
# Make e-mail message
# =============================================================================

def make_email(from_addr: str,
               date: str = None,
               sender: str = "",
               reply_to: Union[str, List[str]] = "",
               to: Union[str, List[str]] = "",
               cc: Union[str, List[str]] = "",
               bcc: Union[str, List[str]] = "",
               subject: str = "",
               body: str = "",
               content_type: str = CONTENT_TYPE_TEXT,
               charset: str = UTF8,
               attachment_filenames: Sequence[str] = None,
               attachment_binaries: Sequence[bytes] = None,
               attachment_binary_filenames: Sequence[str] = None,
               verbose: bool = False) -> email.mime.multipart.MIMEMultipart:
    """
    Makes an e-mail message.

    Arguments that can be multiple e-mail addresses are (a) a single e-mail
    address as a string, or (b) a list of strings (each a single e-mail
    address), or (c) a comma-separated list of multiple e-mail addresses.

    Args:
        from_addr:
            name of the sender for the "From:" field
        date:
            e-mail date in RFC 2822 format, or ``None`` for "now"
        sender:
            name of the sender for the "Sender:" field
        reply_to:
            name of the sender for the "Reply-To:" field

        to:
            e-mail address(es) of the recipients for "To:" field
        cc:
            e-mail address(es) of the recipients for "Cc:" field
        bcc:
            e-mail address(es) of the recipients for "Bcc:" field

        subject:
            e-mail subject
        body:
            e-mail body
        content_type:
            MIME type for body content, default ``text/plain``
        charset:
            character set for body; default ``utf8``

        attachment_filenames:
            filenames of attachments to add
        attachment_binaries:
            binary objects to add as attachments
        attachment_binary_filenames:
            filenames corresponding to ``attachment_binaries``
        verbose: be verbose?

    Returns:
        a :class:`email.mime.multipart.MIMEMultipart`

    Raises:
        :exc:`AssertionError`, :exc:`ValueError`

    """
    def _csv_list_to_list(x: str) -> List[str]:
        stripped = [item.strip() for item in x.split(COMMA)]
        return [item for item in stripped if item]

    def _assert_nocomma(x: Union[str, List[str]]) -> None:
        if isinstance(x, str):
            x = [x]
        for _addr in x:
            assert COMMA not in _addr, (
                f"Commas not allowed in e-mail addresses: {_addr!r}"
            )

    # -------------------------------------------------------------------------
    # Arguments
    # -------------------------------------------------------------------------
    if not date:
        date = email.utils.formatdate(localtime=True)
    assert isinstance(from_addr, str), (
        f"'From:' can only be a single address "
        f"(for Python sendmail, not RFC 2822); was {from_addr!r}"
    )
    _assert_nocomma(from_addr)
    assert isinstance(sender, str), (
        f"'Sender:' can only be a single address; was {sender!r}"
    )
    _assert_nocomma(sender)
    if isinstance(reply_to, str):
        reply_to = [reply_to] if reply_to else []  # type: List[str]
    _assert_nocomma(reply_to)
    if isinstance(to, str):
        to = _csv_list_to_list(to)
    if isinstance(cc, str):
        cc = _csv_list_to_list(cc)
    if isinstance(bcc, str):
        bcc = _csv_list_to_list(bcc)
    assert to or cc or bcc, "No recipients (must have some of: To, Cc, Bcc)"
    _assert_nocomma(to)
    _assert_nocomma(cc)
    _assert_nocomma(bcc)
    attachment_filenames = attachment_filenames or []  # type: List[str]
    assert all(attachment_filenames), (
        f"Missing attachment filenames: {attachment_filenames!r}"
    )
    attachment_binaries = attachment_binaries or []  # type: List[bytes]
    attachment_binary_filenames = attachment_binary_filenames or []  # type: List[str]  # noqa
    assert len(attachment_binaries) == len(attachment_binary_filenames), (
        "If you specify attachment_binaries or attachment_binary_filenames, "
        "they must be iterables of the same length."
    )
    assert all(attachment_binary_filenames), (
        f"Missing filenames for attached binaries: "
        f"{attachment_binary_filenames!r}"
    )

    # -------------------------------------------------------------------------
    # Make message
    # -------------------------------------------------------------------------
    msg = email.mime.multipart.MIMEMultipart()

    # Headers: mandatory
    msg["From"] = from_addr
    msg["Date"] = date
    msg["Subject"] = subject

    # Headers: optional
    if sender:
        msg["Sender"] = sender  # Single only, not a list
    if reply_to:
        msg["Reply-To"] = COMMASPACE.join(reply_to)
    if to:
        msg["To"] = COMMASPACE.join(to)
    if cc:
        msg["Cc"] = COMMASPACE.join(cc)
    if bcc:
        msg["Bcc"] = COMMASPACE.join(bcc)

    # Body
    if content_type == CONTENT_TYPE_TEXT:
        msgbody = email.mime.text.MIMEText(body, "plain", charset)
    elif content_type == CONTENT_TYPE_HTML:
        msgbody = email.mime.text.MIMEText(body, "html", charset)
    else:
        raise ValueError("unknown content_type")
    msg.attach(msgbody)

    # Attachments
    # noinspection PyPep8,PyBroadException
    try:
        if attachment_filenames:
            # -----------------------------------------------------------------
            # Attach things by filename
            # -----------------------------------------------------------------
            if verbose:
                log.debug("attachment_filenames: {}", attachment_filenames)
            # noinspection PyTypeChecker
            for f in attachment_filenames:
                part = email.mime.base.MIMEBase("application", "octet-stream")
                part.set_payload(open(f, "rb").read())
                email.encoders.encode_base64(part)
                part.add_header(
                    'Content-Disposition',
                    'attachment; filename="%s"' % os.path.basename(f)
                )
                msg.attach(part)
        if attachment_binaries:
            # -----------------------------------------------------------------
            # Binary attachments, which have a notional filename
            # -----------------------------------------------------------------
            if verbose:
                log.debug("attachment_binary_filenames: {}",
                          attachment_binary_filenames)
            for i in range(len(attachment_binaries)):
                blob = attachment_binaries[i]
                filename = attachment_binary_filenames[i]
                part = email.mime.base.MIMEBase("application", "octet-stream")
                part.set_payload(blob)
                email.encoders.encode_base64(part)
                part.add_header(
                    'Content-Disposition',
                    'attachment; filename="%s"' % filename)
                msg.attach(part)
    except Exception as e:
        raise ValueError(f"send_email: Failed to attach files: {e}")

    return msg


# =============================================================================
# Send message
# =============================================================================

def send_msg(from_addr: str,
             to_addrs: Union[str, List[str]],
             host: str,
             user: str,
             password: str,
             port: int = None,
             use_tls: bool = True,
             msg: email.mime.multipart.MIMEMultipart = None,
             msg_string: str = None) -> None:
    """
    Sends a pre-built e-mail message.

    Args:
        from_addr: e-mail address for 'From:' field
        to_addrs: address or list of addresses to transmit to

        host: mail server host
        user: username on mail server
        password: password for username on mail server
        port: port to use, or ``None`` for protocol default
        use_tls: use TLS, rather than plain SMTP?

        msg: a :class:`email.mime.multipart.MIMEMultipart`
        msg_string: alternative: specify the message as a raw string

    Raises:
        :exc:`RuntimeError`

    See also:

    - https://tools.ietf.org/html/rfc3207

    """
    assert bool(msg) != bool(msg_string), "Specify either msg or msg_string"
    # Connect
    try:
        session = smtplib.SMTP(host, port)
    except OSError as e:
        # https://bugs.python.org/issue2118
        # Not all errors from smtplib are raised as SMTPException
        # e.g. ConnectionRefusedError when creating the socket
        # SMTPException is a subclass of OSError since 3.4
        raise RuntimeError(
            f"send_msg: Failed to connect to host {host}, port {port}: {e}")
    try:
        session.ehlo()
    except smtplib.SMTPException as e:
        raise RuntimeError(f"send_msg: Failed to issue EHLO: {e}")

    if use_tls:
        try:
            session.starttls()
            session.ehlo()
        except smtplib.SMTPException as e:
            raise RuntimeError(f"send_msg: Failed to initiate TLS: {e}")

    # Log in
    if user:
        try:
            session.login(user, password)
        except smtplib.SMTPException as e:
            raise RuntimeError(
                f"send_msg: Failed to login as user {user}: {e}")
    else:
        log.debug("Not using SMTP AUTH; no user specified")
        # For systems with... lax... security requirements

    # Send
    try:
        session.sendmail(from_addr, to_addrs, msg.as_string())
    except smtplib.SMTPException as e:
        raise RuntimeError(f"send_msg: Failed to send e-mail: {e}")

    # Log out
    session.quit()


# =============================================================================
# Send e-mail
# =============================================================================

def send_email(from_addr: str,
               host: str,
               user: str,
               password: str,
               port: int = None,
               use_tls: bool = True,
               date: str = None,
               sender: str = "",
               reply_to: Union[str, List[str]] = "",
               to: Union[str, List[str]] = "",
               cc: Union[str, List[str]] = "",
               bcc: Union[str, List[str]] = "",
               subject: str = "",
               body: str = "",
               content_type: str = CONTENT_TYPE_TEXT,
               charset: str = UTF8,
               attachment_filenames: Sequence[str] = None,
               attachment_binaries: Sequence[bytes] = None,
               attachment_binary_filenames: Sequence[str] = None,
               verbose: bool = False) -> Tuple[bool, str]:
    """
    Sends an e-mail in text/html format using SMTP via TLS.

    Args:
        host:
            mail server host
        user:
            username on mail server
        password:
            password for username on mail server
        port:
            port to use, or ``None`` for protocol default
        use_tls:
            use TLS, rather than plain SMTP?
        
        date:
            e-mail date in RFC 2822 format, or ``None`` for "now"
        
        from_addr:
            name of the sender for the "From:" field
        sender:
            name of the sender for the "Sender:" field
        reply_to:
            name of the sender for the "Reply-To:" field
        
        to:
            e-mail address(es) of the recipients for "To:" field
        cc:
            e-mail address(es) of the recipients for "Cc:" field
        bcc:
            e-mail address(es) of the recipients for "Bcc:" field
        
        subject:
            e-mail subject
        body:
            e-mail body
        content_type:
            MIME type for body content, default ``text/plain``
        charset:    
            character set for body; default ``utf8``
        
        attachment_filenames:
            filenames of attachments to add
        attachment_binaries:
            binary objects to add as attachments
        attachment_binary_filenames:
            filenames corresponding to ``attachment_binaries``
        verbose:
            be verbose?

    Returns:
         tuple: ``(success, error_or_success_message)``

    See

    - https://tools.ietf.org/html/rfc2822
    - https://tools.ietf.org/html/rfc5322
    - https://segfault.in/2010/12/sending-gmail-from-python/
    - https://stackoverflow.com/questions/64505
    - https://stackoverflow.com/questions/3362600

    Re security:

    - TLS supersedes SSL:
      https://en.wikipedia.org/wiki/Transport_Layer_Security
      
    - https://en.wikipedia.org/wiki/Email_encryption
    
    - SMTP connections on ports 25 and 587 are commonly secured via TLS using
      the ``STARTTLS`` command:
      https://en.wikipedia.org/wiki/Simple_Mail_Transfer_Protocol
      
    - https://tools.ietf.org/html/rfc8314
    
    - "STARTTLS on port 587" is one common method. Django refers to this as
      "explicit TLS" (its ``E_MAIL_USE_TLS`` setting; see
      https://docs.djangoproject.com/en/2.1/ref/settings/#std:setting-EMAIL_USE_TLS).
      
    - Port 465 is also used for "implicit TLS" (3.3 in
      https://tools.ietf.org/html/rfc8314). Django refers to this as "implicit
      TLS" too, or SSL; see its ``EMAIL_USE_SSL`` setting at
      https://docs.djangoproject.com/en/2.1/ref/settings/#email-use-ssl). We
      don't support that here.

    """  # noqa
    if isinstance(to, str):
        to = [to]
    if isinstance(cc, str):
        cc = [cc]
    if isinstance(bcc, str):
        bcc = [bcc]

    # -------------------------------------------------------------------------
    # Make it
    # -------------------------------------------------------------------------
    try:
        msg = make_email(
            from_addr=from_addr,
            date=date,
            sender=sender,
            reply_to=reply_to,
            to=to,
            cc=cc,
            bcc=bcc,
            subject=subject,
            body=body,
            content_type=content_type,
            charset=charset,
            attachment_filenames=attachment_filenames,
            attachment_binaries=attachment_binaries,
            attachment_binary_filenames=attachment_binary_filenames,
            verbose=verbose,
        )
    except (AssertionError, ValueError) as e:
        errmsg = str(e)
        log.error("{}", errmsg)
        return False, errmsg

    # -------------------------------------------------------------------------
    # Send it
    # -------------------------------------------------------------------------

    to_addrs = to + cc + bcc
    try:
        send_msg(
            msg=msg,
            from_addr=from_addr,
            to_addrs=to_addrs,
            host=host,
            user=user,
            password=password,
            port=port,
            use_tls=use_tls,
        )
    except RuntimeError as e:
        errmsg = str(e)
        log.error("{}", e)
        return False, errmsg

    return True, "Success"


# =============================================================================
# Misc
# =============================================================================

_SIMPLE_EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")


def is_email_valid(email_: str) -> bool:
    """
    Performs some basic checks that a string appears to be an e-mail address.

    See 
    https://stackoverflow.com/questions/8022530/how-to-check-for-valid-email-address.
    """  # noqa
    # Very basic checks!
    if not email_:
        return False
    if _SIMPLE_EMAIL_REGEX.match(email_) is None:
        return False
    # The other things that can get through:
    # Exclude e.g. two @ symbols
    if email_.count("@") != 1:
        return False
    # Commas are not allowed
    if email_.count(",") != 0:
        return False
    # Can't end in a full stop:
    if email_.endswith("."):
        return False
    return True


class TestIsEmailValid(unittest.TestCase):
    """
    Run with e.g. ``pytest sendmail.py::TestIsEmailValid``
    """
    def test_is_email_valid(self) -> None:
        good_addresses = [
            "x@somewhere.com",
            "a+b@somewhere.else",
            "fish123@blah.com",
        ]
        bad_addresses = [
            "xyz",
            "thing@blah.com@blah.com",
            "thing,with_comma@somewhere.co.uk",
            "person@place.fullstop.",
        ]
        for good in good_addresses:
            self.assertTrue(
                is_email_valid(good),
                f"Good e-mail being flagged as bad: {good!r}"
            )
        for bad in bad_addresses:
            self.assertFalse(
                is_email_valid(bad),
                f"Bad e-mail being flagged as good: {bad!r}"
            )


def get_email_domain(email_: str) -> str:
    """
    Returns the domain part of an e-mail address.
    """
    return email_.split("@")[1]


# =============================================================================
# Parse command line
# =============================================================================

def main() -> NoReturn:
    """
    Command-line processor. See ``--help`` for details.
    """
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    parser = argparse.ArgumentParser(
        description="Send an e-mail from the command line.")
    parser.add_argument("sender", action="store",
                        help="Sender's e-mail address")
    parser.add_argument("host", action="store",
                        help="SMTP server hostname")
    parser.add_argument("user", action="store",
                        help="SMTP username")
    parser.add_argument("password", action="store",
                        help="SMTP password")
    parser.add_argument("recipient", action="append",
                        help="Recipient e-mail address(es)")
    parser.add_argument("subject", action="store",
                        help="Message subject")
    parser.add_argument("body", action="store",
                        help="Message body")
    parser.add_argument("--attach", nargs="*",
                        help="Filename(s) to attach")
    parser.add_argument("--tls", action="store_false",
                        help="Use TLS connection security")
    parser.add_argument("--verbose", action="store_true",
                        help="Be verbose")
    parser.add_argument("-h --help", action="help",
                        help="Prints this help")
    args = parser.parse_args()
    (result, msg) = send_email(
        from_addr=args.sender,
        to=args.recipient,
        subject=args.subject,
        body=args.body,
        host=args.host,
        user=args.user,
        password=args.password,
        use_tls=args.tls,
        attachment_filenames=args.attach,
        verbose=args.verbose,
    )
    if result:
        log.info("Success")
    else:
        log.info("Failure")
        # log.error(msg)
    sys.exit(0 if result else 1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/plot.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support for plotting via matplotlib.pyplot.**
"""

import io
# noinspection PyUnresolvedReferences
from types import ModuleType
from typing import TYPE_CHECKING, Union

from cardinal_pythonlib import rnc_web

if TYPE_CHECKING:
    try:
        # noinspection PyPackageRequirements
        from matplotlib.figure import Figure
    except ImportError:
        Figure = None


# =============================================================================
# Image embedding in PDFs
# =============================================================================
# xhtml2pdf (2013-04-11) supports PNG, but not SVG.
# You can convert SVG to PNG for embedding:
# https://stackoverflow.com/questions/787287
# You could make a PDF and append it, though that would (without further
# effort) lack the patient headers.

def png_img_html_from_pyplot_figure(fig: "Figure",
                                    dpi: int = 100,
                                    extra_html_class: str = None) -> str:
    """
    Converts a ``pyplot`` figure to an HTML IMG tag with encapsulated PNG.
    """
    if fig is None:
        return ""
    # Make a file-like object
    memfile = io.BytesIO()
    # In general, can do
    #   fig.savefig(filename/file-like-object/backend, format=...)
    # or
    #   backend.savefig(fig):
    # see e.g. http://matplotlib.org/api/backend_pdf_api.html
    # noinspection PyUnresolvedReferences
    fig.savefig(memfile, format="png", dpi=dpi)
    memfile.seek(0)
    pngblob = memoryview(memfile.read())
    return rnc_web.get_png_img_html(pngblob, extra_html_class)


def svg_html_from_pyplot_figure(fig: "Figure") -> str:
    """
    Converts a ``pyplot`` figure to an SVG tag.
    """
    if fig is None:
        return ""
    memfile = io.BytesIO()  # StringIO doesn't like mixing str/unicode
    # noinspection PyUnresolvedReferences
    fig.savefig(memfile, format="svg")
    return memfile.getvalue().decode("utf-8")  # returns a text/Unicode type
    # SVG works directly in HTML; it returns <svg ...></svg>


# =============================================================================
# Plotting
# =============================================================================

def set_matplotlib_fontsize(matplotlib: ModuleType,
                            fontsize: Union[int, float] = 12) -> None:
    """
    Sets the current font size within the ``matplotlib`` library.

    **WARNING:** not an appropriate method for multithreaded environments, as
    it writes (indirectly) to ``matplotlib`` global objects. See CamCOPS for
    alternative methods.
    """
    font = {
        # https://stackoverflow.com/questions/3899980
        # http://matplotlib.org/users/customizing.html
        'family': 'sans-serif',
        # ... serif, sans-serif, cursive, fantasy, monospace
        'style': 'normal',  # normal (roman), italic, oblique
        'variant': 'normal',  # normal, small-caps
        'weight': 'normal',
        # ... normal [=400], bold [=700], bolder [relative to current],
        # lighter [relative], 100, 200, 300, ..., 900
        'size': fontsize  # in pt (default 12)
    }
    # noinspection PyUnresolvedReferences
    matplotlib.rc('font', **font)
    legend = {
        # https://stackoverflow.com/questions/7125009
        'fontsize': fontsize
    }
    # noinspection PyUnresolvedReferences
    matplotlib.rc('legend', **legend)



#!/usr/bin/env python
# cardinal_pythonlib/cmdline.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions for manipulating command-line parameters.**

"""

import re
# import shlex
import subprocess
import sys
from typing import List, Union


def cmdline_split(s: str, platform: Union[int, str] = 'this') -> List[str]:
    """
    As per
    https://stackoverflow.com/questions/33560364/python-windows-parsing-command-lines-with-shlex.

    Multi-platform variant of ``shlex.split()`` for command-line splitting.
    For use with ``subprocess``, for ``argv`` injection etc. Using fast REGEX.
    
    Args:
        s:
            string to split
        platform:
            - ``'this'`` = auto from current platform;
            - ``1`` = POSIX;
            - ``0`` = Windows/CMD
            - (other values reserved)
    """  # noqa
    if platform == 'this':
        platform = (sys.platform != 'win32')  # RNC: includes 64-bit Windows

    if platform == 1:  # POSIX
        re_cmd_lex = r'''"((?:\\["\\]|[^"])*)"|'([^']*)'|(\\.)|(&&?|\|\|?|\d?\>|[<])|([^\s'"\\&|<>]+)|(\s+)|(.)'''  # noqa
    elif platform == 0:  # Windows/CMD
        re_cmd_lex = r'''"((?:""|\\["\\]|[^"])*)"?()|(\\\\(?=\\*")|\\")|(&&?|\|\|?|\d?>|[<])|([^\s"&|<>]+)|(\s+)|(.)'''  # noqa
    else:
        raise AssertionError(f"unknown platform {platform!r}")

    args = []
    accu = None   # collects pieces of one arg
    for qs, qss, esc, pipe, word, white, fail in re.findall(re_cmd_lex, s):
        if word:
            pass   # most frequent
        elif esc:
            word = esc[1]
        elif white or pipe:
            if accu is not None:
                args.append(accu)
            if pipe:
                args.append(pipe)
            accu = None
            continue
        elif fail:
            raise ValueError("invalid or incomplete shell string")
        elif qs:
            word = qs.replace(r'\"', '"').replace(r'\\', '\\')
            # ... raw strings can't end in single backslashes;
            # https://stackoverflow.com/questions/647769/why-cant-pythons-raw-string-literals-end-with-a-single-backslash  # noqa
            if platform == 0:
                word = word.replace('""', '"')
        else:
            word = qss   # may be even empty; must be last

        accu = (accu or '') + word

    if accu is not None:
        args.append(accu)

    return args


def cmdline_quote_posix(seq: List[str]) -> str:
    """
    Quotes arguments for POSIX, producing a single string suitable for
    copying/pasting.

    Based on subprocess.list2cmdline().
    """
    result = []  # type: List[str]
    for arg in seq:
        bs_buf = []  # type: List[str]

        # Add a space to separate this argument from the others
        if result:
            result.append(' ')

        # Modified here: quote arguments with "*"
        needquote = (" " in arg) or ("\t" in arg) or ("*" in arg) or not arg
        if needquote:
            result.append('"')

        for c in arg:
            if c == '\\':
                # Don't know if we need to double yet.
                bs_buf.append(c)
            elif c == '"':
                # Double backslashes.
                result.append('\\' * len(bs_buf) * 2)
                bs_buf = []
                result.append('\\"')
            else:
                # Normal char
                if bs_buf:
                    result.extend(bs_buf)
                    bs_buf = []
                result.append(c)

        # Add remaining backslashes, if any.
        if bs_buf:
            result.extend(bs_buf)

        if needquote:
            result.extend(bs_buf)
            result.append('"')

    return ''.join(result)


def cmdline_quote(args: List[str], platform: Union[int, str] = 'this') -> str:
    """
    Convert a list of command-line arguments to a suitably quoted command-line
    string that should be copy/pastable into a comand prompt.
    """
    if platform == 'this':
        platform = (sys.platform != 'win32')  # RNC: includes 64-bit Windows

    if platform == 1:  # POSIX
        return cmdline_quote_posix(args)
    elif platform == 0:  # Windows/CMD
        return subprocess.list2cmdline(args)
    else:
        raise AssertionError(f"unknown platform {platform!r}")



#!/usr/bin/env python3
# cardinal_pythonlib/openxml/find_bad_openxml.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Tool to scan rescued Microsoft Office OpenXML files (produced by the
"find_recovered_openxml.py" tool in this kit; q.v.) and detect bad
(corrupted) ones.**

"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import fnmatch
import logging
import multiprocessing
# from pprint import pformat
import os
from sys import stdin
from time import sleep
import traceback
from typing import Generator, List
from zipfile import BadZipFile, ZipFile, ZipInfo

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)
from cardinal_pythonlib.fileops import exists_locked, gen_filenames
from cardinal_pythonlib.openxml.find_recovered_openxml import (
    DOCX_CONTENTS_REGEX,
    PPTX_CONTENTS_REGEX,
    XLSX_CONTENTS_REGEX,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

MANDATORY_FILENAMES = [
    # https://msdn.microsoft.com/en-us/library/aa982683(v=office.12).aspx

    "[Content_Types].xml",

    # "_rels/.rels",
    # ... not strictly mandatory:
    #     https://en.wikipedia.org/wiki/Open_Packaging_Conventions

    # "docProps/core.xml"
    # ... NOT mandatory, or at least I have a .pptx that works fine without it
]

NULL_DATE_TIME = (1980, 1, 1, 0, 0, 0)


def gen_from_stdin() -> Generator[str, None, None]:
    """
    Yields stripped lines from stdin.
    """
    for line in stdin.readlines():
        yield line.strip()


def is_openxml_good(filename: str) -> bool:
    """
    Determines whether an OpenXML file appears to be good (not corrupted).
    """
    try:
        log.debug("Trying: {}", filename)
        with ZipFile(filename, 'r') as zip_ref:
            namelist = zip_ref.namelist()  # type: List[str]
            # log.critical("\n{}", pformat(namelist))
            # -----------------------------------------------------------------
            # Contains key files?
            # -----------------------------------------------------------------
            for mandatory_filename in MANDATORY_FILENAMES:
                if mandatory_filename not in namelist:
                    log.debug("Bad [missing {!r}]: {}",
                              mandatory_filename, filename)
                    return False

            infolist = zip_ref.infolist()  # type: List[ZipInfo]
            contains_docx = False
            contains_pptx = False
            contains_xlsx = False
            for info in infolist:
                # -------------------------------------------------------------
                # Sensible date check?
                # ... NO: lots of perfectly good files have this date/time.
                # -------------------------------------------------------------
                # if info.date_time == NULL_DATE_TIME:
                #     log.debug("{!r}: {!r}", info.filename, info.date_time)

                # -------------------------------------------------------------
                # Only one kind of contents?
                # ... YES, I think so. This has 100% reliability on my
                # stash of 34 PPTX, 223 DOCX, 85 XLSX, and labelled none as bad
                # from an HFC collection of 1866 such files. There are lots of
                # files emerging from Scalpel (plus my find_recovered_openxml
                # zip-fixing tool) that fail this test, though.
                # -------------------------------------------------------------
                if (not contains_docx and
                        DOCX_CONTENTS_REGEX.search(info.filename)):
                    contains_docx = True
                if (not contains_pptx and
                        PPTX_CONTENTS_REGEX.search(info.filename)):
                    contains_pptx = True
                if (not contains_xlsx and
                        XLSX_CONTENTS_REGEX.search(info.filename)):
                    contains_xlsx = True
                if sum([contains_docx, contains_pptx, contains_xlsx]) > 1:
                    log.debug("Bad [>1 of DOCX, PPTX, XLSX content]: {}",
                              filename)
                    return False

            return True
    except (BadZipFile, OSError) as e:
        # ---------------------------------------------------------------------
        # Duff file. Easy!
        # ---------------------------------------------------------------------
        log.debug("Bad [BadZipFile or OSError]: {!r}; error was {!r}",
                  filename, e)
        return False


def process_openxml_file(filename: str,
                         print_good: bool,
                         delete_if_bad: bool) -> None:
    """
    Prints the filename of, or deletes, an OpenXML file depending on whether
    it is corrupt or not.

    Args:
        filename: filename to check
        print_good: if ``True``, then prints the filename if the file
            appears good.
        delete_if_bad: if ``True``, then deletes the file if the file
            appears corrupt.
    """
    print_bad = not print_good
    try:
        file_good = is_openxml_good(filename)
        file_bad = not file_good
        if (print_good and file_good) or (print_bad and file_bad):
            print(filename)
        if delete_if_bad and file_bad:
            log.warning("Deleting: {}", filename)
            os.remove(filename)
    except Exception as e:
        # Must explicitly catch and report errors, since otherwise they vanish
        # into the ether.
        log.critical("Uncaught error in subprocess: {!r}\n{}", e,
                     traceback.format_exc())
        raise


def main() -> None:
    """
    Command-line handler for the ``find_bad_openxml`` tool.
    Use the ``--help`` option for help.
    """
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description="""
Tool to scan rescued Microsoft Office OpenXML files (produced by the
find_recovered_openxml.py tool in this kit; q.v.) and detect bad (corrupted)
ones.
        """
    )
    parser.add_argument(
        "filename", nargs="*",
        help="File(s) to check. You can also specify directores if you use "
             "--recursive"
    )
    parser.add_argument(
        "--filenames_from_stdin", "-x", action="store_true",
        help="Take filenames from stdin instead, one line per filename "
             "(useful for chained grep)."
    )
    parser.add_argument(
        "--recursive", action="store_true",
        help="Allow search to descend recursively into any directories "
             "encountered."
    )
    parser.add_argument(
        "--skip_files", nargs="*", default=[],
        help="File pattern(s) to skip. You can specify wildcards like '*.txt' "
             "(but you will have to enclose that pattern in quotes under "
             "UNIX-like operating systems). The basename of each file will be "
             "tested against these filenames/patterns. Consider including "
             "Scalpel's 'audit.txt'."
    )
    parser.add_argument(
        "--good", action="store_true",
        help="List good files, not bad"
    )
    parser.add_argument(
        "--delete_if_bad", action="store_true",
        help="If a file is found to be bad, delete it. DANGEROUS."
    )
    parser.add_argument(
        "--run_repeatedly", type=int,
        help="Run the tool repeatedly with a pause of <run_repeatedly> "
             "seconds between runs. (For this to work well with the move/"
             "delete options, you should specify one or more DIRECTORIES in "
             "the 'filename' arguments, not files, and you will need the "
             "--recursive option.)"
    )
    parser.add_argument(
        "--nprocesses", type=int, default=multiprocessing.cpu_count(),
        help="Specify the number of processes to run in parallel."
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger(
        level=logging.DEBUG if args.verbose else logging.INFO,
        with_process_id=True
    )
    if bool(args.filenames_from_stdin) == bool(args.filename):
        raise ValueError("Specify --filenames_from_stdin or filenames on the "
                         "command line, but not both")
    if args.filenames_from_stdin and args.run_repeatedly:
        raise ValueError("Can't use both --filenames_from_stdin and "
                         "--run_repeatedly")

    # Repeated scanning loop
    while True:
        log.debug("Starting scan.")
        log.debug("- Scanning files/directories {!r}{}",
                  args.filename,
                  " recursively" if args.recursive else "")
        log.debug("- Skipping files matching {!r}", args.skip_files)
        log.debug("- Using {} simultaneous processes", args.nprocesses)
        log.debug("- Reporting {} filenames", "good" if args.good else "bad")
        if args.delete_if_bad:
            log.warning("- Deleting bad OpenXML files.")

        # Iterate through files
        pool = multiprocessing.Pool(processes=args.nprocesses)

        if args.filenames_from_stdin:
            generator = gen_from_stdin()
        else:
            generator = gen_filenames(starting_filenames=args.filename,
                                      recursive=args.recursive)

        for filename in generator:
            src_basename = os.path.basename(filename)
            if any(fnmatch.fnmatch(src_basename, pattern)
                   for pattern in args.skip_files):
                log.debug("Skipping file as ordered: " + filename)
                continue
            exists, locked = exists_locked(filename)
            if locked or not exists:
                log.debug("Skipping currently inaccessible file: " + filename)
                continue
            kwargs = {
                'filename': filename,
                'print_good': args.good,
                'delete_if_bad': args.delete_if_bad,
            }
            # log.critical("start")
            pool.apply_async(process_openxml_file, [], kwargs)
            # result = pool.apply_async(process_file, [], kwargs)
            # result.get()  # will re-raise any child exceptions
            # ... but it waits for the process to complete! That's no help.
            # log.critical("next")
            # ... https://stackoverflow.com/questions/22094852/how-to-catch-exceptions-in-workers-in-multiprocessing  # noqa
        pool.close()
        pool.join()

        log.debug("Finished scan.")
        if args.run_repeatedly is None:
            break
        log.info("Sleeping for {} s...", args.run_repeatedly)
        sleep(args.run_repeatedly)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/metaclasses.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions to do with metaclasses. IGNORE; NOT WORKING PROPERLY.**

"""

from typing import Any, Dict, Tuple, Type


class CooperativeMeta(type):
    """
    The idea is to use this as the metaclass for a class ``Derived`` that
    inherits from bases ``BaseOne``, ``BaseTwo``, ..., whose metaclasses are
    not of the same type.

    This should avoid the error:

    .. code-block:: none

        TypeError: metaclass conflict: the metaclass of a derived class must
        be a (non-strict) subclass of the metaclasses of all its bases

    Code from:
    https://stackoverflow.com/questions/6557407/triple-inheritance-causes-metaclass-conflict-sometimes.

    See also:
    https://code.activestate.com/recipes/204197-solving-the-metaclass-conflict/.

    HOWEVER, it's not actually working. Does Python 3 have stricter checking
    than Python 2?

    See also
    https://blog.ionelmc.ro/2015/02/09/understanding-python-metaclasses/.
    """  # noqa
    def __new__(mcs: Type,
                name: str,
                bases: Tuple[Type, ...],
                members: Dict[str, Any]) -> Type:
        # collect up the metaclasses
        metas = [type(base) for base in bases]

        # prune repeated or conflicting entries
        metas = [meta for index, meta in enumerate(metas)
                 if not [later for later in metas[index + 1:]
                         if issubclass(later, meta)]]

        # whip up the actual combined meta class; derive off all of these
        meta = type(name, tuple(metas), dict(combined_metas=metas))

        # make the actual object
        return meta(name, bases, members)

    # noinspection PyMissingConstructor
    def __init__(cls: Type,
                 name: str,
                 bases: Tuple[Type, ...],
                 members: Dict[str, Any]) -> None:
        # noinspection PyUnresolvedReferences
        for meta in cls.combined_metas:
            meta.__init__(cls, name, bases, members)


class DebuggingCooperativeMeta(type):
    """
    ``CooperativeMeta``, but with :func:`print` output.
    Still not working.
    """
    def __new__(mcs: Type,
                name: str,
                bases: Tuple[Type, ...],
                members: Dict[str, Any]) -> Type:
        metas = [type(base) for base in bases]
        metas = [meta for index, meta in enumerate(metas)
                 if not [later for later in metas[index + 1:]
                         if issubclass(later, meta)]]
        meta = type(name, tuple(metas), dict(combined_metas=metas))
        obj = meta(name, bases, members)

        print(f"DebuggingCooperativeMeta.__new__: bases = {bases!r}")
        print(f"DebuggingCooperativeMeta.__new__: metas = {metas!r}")
        print(f"DebuggingCooperativeMeta.__new__: meta = {meta!r}")
        print(f"DebuggingCooperativeMeta.__new__: obj = {obj!r}")

        # make the actual object
        return obj

    # noinspection PyMissingConstructor
    def __init__(cls: Type,
                 name: str,
                 bases: Tuple[Type, ...],
                 members: Dict[str, Any]) -> None:
        # noinspection PyUnresolvedReferences
        for meta in cls.combined_metas:
            print(
                f"DebuggingCooperativeMeta.__init__: calling "
                f"{meta.__name__}.__init__({name!r}, {bases!r}, {members!r}"
            )
            meta.__init__(cls, name, bases, members)


TEST_CODE_SHOULD_FAIL = """

class MetaA(type):
    pass

class MetaB(type):
    pass

class A(metaclass=MetaA):
    pass

class B(metaclass=MetaB):
    pass

class Fixed(A, B):
    pass

"""

TEST_CODE_SHOULD_SUCCEED_BUT_DOES_NOT = """

from cardinal_pythonlib.metaclasses import CooperativeMeta

class MetaA(type):
    pass

class MetaB(type):
    pass

class A(metaclass=MetaA):
    pass

class B(metaclass=MetaB):
    pass

class Fixed(A, B, metaclass=CooperativeMeta):
    pass

"""
